import abc
import numbers
from collections import OrderedDict, defaultdict

import numpy as np

from .utils import deserialize_class

__all__ = [
    "BaseHighLevelWCS",
    "HighLevelWCSMixin",
    "high_level_objects_to_values",
    "values_to_high_level_objects",
]


def rec_getattr(obj, att):
    for a in att.split("."):
        obj = getattr(obj, a)
    return obj


def default_order(components):
    order = []
    for key, _, _ in components:
        if key not in order:
            order.append(key)
    return order


def _toindex(value):
    """Convert value to an int or an int array.

    Input coordinates converted to integers
    corresponding to the center of the pixel.
    The convention is that the center of the pixel is
    (0, 0), while the lower left corner is (-0.5, -0.5).
    The outputs are used to index the mask.

    Examples
    --------
    >>> _toindex(np.array([-0.5, 0.49999]))
    array([0, 0])
    >>> _toindex(np.array([0.5, 1.49999]))
    array([1, 1])
    >>> _toindex(np.array([1.5, 2.49999]))
    array([2, 2])
    """
    arr = np.floor(np.asarray(value) + 0.5)

    fill_value = np.iinfo(int).min
    if np.isscalar(arr):
        if np.isnan(arr):
            arr = fill_value
    else:
        arr[np.isnan(arr)] = fill_value
    return np.asarray(arr, dtype=int)


class BaseHighLevelWCS(metaclass=abc.ABCMeta):
    """
    Abstract base class for the high-level WCS interface.

    This is described in `APE 14: A shared Python interface for World Coordinate
    Systems <https://doi.org/10.5281/zenodo.1188875>`_.
    """

    @property
    @abc.abstractmethod
    def low_level_wcs(self):
        """
        Returns a reference to the underlying low-level WCS object.
        """

    @abc.abstractmethod
    def pixel_to_world(self, *pixel_arrays):
        """
        Convert pixel coordinates to world coordinates (represented by
        high-level objects).

        If a single high-level object is used to represent the world coordinates
        (i.e., if ``len(wcs.world_axis_object_classes) == 1``), it is returned
        as-is (not in a tuple/list), otherwise a tuple of high-level objects is
        returned. See
        `~astropy.wcs.wcsapi.BaseLowLevelWCS.pixel_to_world_values` for pixel
        indexing and ordering conventions.
        """

    def array_index_to_world(self, *index_arrays):
        """
        Convert array indices to world coordinates (represented by Astropy
        objects).

        If a single high-level object is used to represent the world coordinates
        (i.e., if ``len(wcs.world_axis_object_classes) == 1``), it is returned
        as-is (not in a tuple/list), otherwise a tuple of high-level objects is
        returned. See
        `~astropy.wcs.wcsapi.BaseLowLevelWCS.array_index_to_world_values` for
        pixel indexing and ordering conventions.
        """
        return self.pixel_to_world(*index_arrays[::-1])

    @abc.abstractmethod
    def world_to_pixel(self, *world_objects):
        """
        Convert world coordinates (represented by Astropy objects) to pixel
        coordinates.

        If `~astropy.wcs.wcsapi.BaseLowLevelWCS.pixel_n_dim` is ``1``, this
        method returns a single scalar or array, otherwise a tuple of scalars or
        arrays is returned. See
        `~astropy.wcs.wcsapi.BaseLowLevelWCS.world_to_pixel_values` for pixel
        indexing and ordering conventions.
        """

    def world_to_array_index(self, *world_objects):
        """
        Convert world coordinates (represented by Astropy objects) to array
        indices.

        If `~astropy.wcs.wcsapi.BaseLowLevelWCS.pixel_n_dim` is ``1``, this
        method returns a single scalar or array, otherwise a tuple of scalars or
        arrays is returned. See
        `~astropy.wcs.wcsapi.BaseLowLevelWCS.world_to_array_index_values` for
        pixel indexing and ordering conventions. The indices should be returned
        as rounded integers.
        """
        if self.low_level_wcs.pixel_n_dim == 1:
            return _toindex(self.world_to_pixel(*world_objects))
        else:
            return tuple(
                np.asarray(x)
                for x in _toindex(self.world_to_pixel(*world_objects)[::-1])
            )


def high_level_objects_to_values(*world_objects, low_level_wcs):
    """
    Convert the input high level object to low level values.

    This function uses the information in ``wcs.world_axis_object_classes`` and
    ``wcs.world_axis_object_components`` to convert the high level objects
    (such as `~.SkyCoord`) to low level "values" which should be scalars or
    Numpy arrays.

    This is used in `.HighLevelWCSMixin.world_to_pixel`, but provided as a
    separate function for use in other places where needed.

    Parameters
    ----------
    *world_objects: object
        High level coordinate objects.

    low_level_wcs: `.BaseLowLevelWCS`
        The WCS object to use to interpret the coordinates.
    """
    # Cache the classes and components since this may be expensive
    serialized_classes = low_level_wcs.world_axis_object_classes
    components = low_level_wcs.world_axis_object_components

    # Deserialize world_axis_object_classes using the default order
    classes = OrderedDict()
    for key in default_order(components):
        if low_level_wcs.serialized_classes:
            classes[key] = deserialize_class(serialized_classes[key], construct=False)
        else:
            classes[key] = serialized_classes[key]

    # Check that the number of classes matches the number of inputs
    if len(world_objects) != len(classes):
        raise ValueError(
            f"Number of world inputs ({len(world_objects)}) does not match expected"
            f" ({len(classes)})"
        )

    # Determine whether the classes are uniquely matched, that is we check
    # whether there is only one of each class.
    world_by_key = {}
    unique_match = True
    for w in world_objects:
        matches = []
        for key, (klass, *_) in classes.items():
            if isinstance(w, klass):
                matches.append(key)
        if len(matches) == 1:
            world_by_key[matches[0]] = w
        else:
            unique_match = False
            break

    # If the match is not unique, the order of the classes needs to match,
    # whereas if all classes are unique, we can still intelligently match
    # them even if the order is wrong.

    objects = {}

    if unique_match:
        for key, (klass, args, kwargs, *rest) in classes.items():
            if len(rest) == 0:
                klass_gen = klass
            elif len(rest) == 1:
                klass_gen = rest[0]
            else:
                raise ValueError(
                    "Tuples in world_axis_object_classes should have length 3 or 4"
                )

            # FIXME: For now SkyCoord won't auto-convert upon initialization
            # https://github.com/astropy/astropy/issues/7689
            from astropy.coordinates import SkyCoord

            if isinstance(world_by_key[key], SkyCoord):
                if "frame" in kwargs:
                    objects[key] = world_by_key[key].transform_to(kwargs["frame"])
                else:
                    objects[key] = world_by_key[key]
            else:
                objects[key] = klass_gen(world_by_key[key], *args, **kwargs)

    else:
        for ikey, key in enumerate(classes):
            klass, args, kwargs, *rest = classes[key]

            if len(rest) == 0:
                klass_gen = klass
            elif len(rest) == 1:
                klass_gen = rest[0]
            else:
                raise ValueError(
                    "Tuples in world_axis_object_classes should have length 3 or 4"
                )

            w = world_objects[ikey]
            if not isinstance(w, klass):
                raise ValueError(
                    "Expected the following order of world arguments:"
                    f" {', '.join([k.__name__ for (k, *_) in classes.values()])}"
                )

            # FIXME: For now SkyCoord won't auto-convert upon initialization
            # https://github.com/astropy/astropy/issues/7689
            from astropy.coordinates import SkyCoord

            if isinstance(w, SkyCoord):
                if "frame" in kwargs:
                    objects[key] = w.transform_to(kwargs["frame"])
                else:
                    objects[key] = w
            else:
                objects[key] = klass_gen(w, *args, **kwargs)

    # We now extract the attributes needed for the world values
    world = []
    for key, _, attr in components:
        if callable(attr):
            world.append(attr(objects[key]))
        else:
            world.append(rec_getattr(objects[key], attr))

    # Check the type of the return values - should be scalars or plain Numpy
    # arrays, not e.g. Quantity. Note that we deliberately use type(w) because
    # we don't want to match Numpy subclasses.
    for w in world:
        if not isinstance(w, numbers.Number) and not type(w) == np.ndarray:
            raise TypeError(
                f"WCS world_axis_object_components results in "
                f"values which are not scalars or plain Numpy "
                f"arrays (got {type(w)})"
            )

    return world


def values_to_high_level_objects(*world_values, low_level_wcs):
    """
    Convert low level values into high level objects.

    This function uses the information in ``wcs.world_axis_object_classes`` and
    ``wcs.world_axis_object_components`` to convert low level "values"
    `~.Quantity` objects, to high level objects (such as `~.SkyCoord`).

    This is used in `.HighLevelWCSMixin.pixel_to_world`, but provided as a
    separate function for use in other places where needed.

    Parameters
    ----------
    *world_values: object
        Low level, "values" representations of the world coordinates.

    low_level_wcs: `.BaseLowLevelWCS`
        The WCS object to use to interpret the coordinates.
    """
    # Check the type of the input values - should be scalars or plain Numpy
    # arrays, not e.g. Quantity. Note that we deliberately use type(w) because
    # we don't want to match Numpy subclasses.
    for w in world_values:
        if not isinstance(w, numbers.Number) and not type(w) == np.ndarray:
            raise TypeError(
                f"Expected world coordinates as scalars or plain Numpy "
                f"arrays (got {type(w)})"
            )

    # Cache the classes and components since this may be expensive
    components = low_level_wcs.world_axis_object_components
    classes = low_level_wcs.world_axis_object_classes

    # Deserialize classes
    if low_level_wcs.serialized_classes:
        classes_new = {}
        for key, value in classes.items():
            classes_new[key] = deserialize_class(value, construct=False)
        classes = classes_new

    args = defaultdict(list)
    kwargs = defaultdict(dict)

    for i, (key, attr, _) in enumerate(components):
        if isinstance(attr, str):
            kwargs[key][attr] = world_values[i]
        else:
            while attr > len(args[key]) - 1:
                args[key].append(None)
            args[key][attr] = world_values[i]

    result = []

    for key in default_order(components):
        klass, ar, kw, *rest = classes[key]
        if len(rest) == 0:
            klass_gen = klass
        elif len(rest) == 1:
            klass_gen = rest[0]
        else:
            raise ValueError(
                "Tuples in world_axis_object_classes should have length 3 or 4"
            )
        result.append(klass_gen(*args[key], *ar, **kwargs[key], **kw))

    return result


class HighLevelWCSMixin(BaseHighLevelWCS):
    """
    Mix-in class that automatically provides the high-level WCS API for the
    low-level WCS object given by the `~HighLevelWCSMixin.low_level_wcs`
    property.
    """

    @property
    def low_level_wcs(self):
        return self

    def world_to_pixel(self, *world_objects):
        world_values = high_level_objects_to_values(
            *world_objects, low_level_wcs=self.low_level_wcs
        )

        # Finally we convert to pixel coordinates
        pixel_values = self.low_level_wcs.world_to_pixel_values(*world_values)

        return pixel_values

    def pixel_to_world(self, *pixel_arrays):
        # Compute the world coordinate values
        world_values = self.low_level_wcs.pixel_to_world_values(*pixel_arrays)

        if self.low_level_wcs.world_n_dim == 1:
            world_values = (world_values,)

        pixel_values = values_to_high_level_objects(
            *world_values, low_level_wcs=self.low_level_wcs
        )

        if len(pixel_values) == 1:
            return pixel_values[0]
        else:
            return pixel_values



# Licensed under a 3-clause BSD style license - see LICENSE.rst


from astropy.cosmology._src.parameter import Parameter, all_parameters


def test_all_parameters():
    """Test :func:`astropy.cosmology._src.utils.all_parameters`."""

    class ClassA:
        a = 1
        b = 2

    got = all_parameters(ClassA)
    assert got == {}

    class ClassB(ClassA):
        H0: Parameter = Parameter(
            doc="Hubble constant at z=0.",
            unit="km/(s Mpc)",
            fvalidate="scalar",
        )

    got = all_parameters(ClassB)
    assert got.keys() == {"H0"}
    assert all(isinstance(p, Parameter) for p in got.values())



# Licensed under a 3-clause BSD style license - see PYFITS.rst

import io
import os
import pathlib
import warnings

import numpy as np
import pytest
from numpy.testing import assert_array_equal

from astropy import units as u
from astropy.io import fits
from astropy.io.fits import printdiff
from astropy.io.fits.connect import REMOVE_KEYWORDS
from astropy.io.fits.tests.test_table import _assert_attr_col
from astropy.table import Table
from astropy.utils.exceptions import AstropyUserWarning

from .conftest import FitsTestCase


class TestConvenience(FitsTestCase):
    def test_resource_warning(self):
        warnings.simplefilter("always", ResourceWarning)
        _ = fits.getdata(self.data("test0.fits"))
        _ = fits.getheader(self.data("test0.fits"))

    def test_fileobj_not_closed(self):
        """
        Tests that file-like objects are not closed after being passed
        to convenience functions.

        Regression test for https://github.com/astropy/astropy/issues/5063
        """

        f = open(self.data("test0.fits"), "rb")
        _ = fits.getdata(f)
        assert not f.closed

        f.seek(0)
        _ = fits.getheader(f)
        assert not f.closed

        f.close()  # Close it now

    def test_table_to_hdu(self):
        table = Table(
            [[1, 2, 3], ["a", "b", "c"], [2.3, 4.5, 6.7]],
            names=["a", "b", "c"],
            dtype=["i", "U1", "f"],
        )
        table["a"].unit = "m/s"
        table["b"].unit = "not-a-unit"

        with pytest.warns(
            u.UnitsWarning, match="'not-a-unit' did not parse as fits unit"
        ) as w:
            hdu = fits.table_to_hdu(table, name="MYTABLE")

        assert len(w) == 1
        assert hdu.header["EXTNAME"] == "MYTABLE"

        # Check that TUNITn cards appear in the correct order
        # (https://github.com/astropy/astropy/pull/5720)
        assert hdu.header.index("TUNIT1") < hdu.header.index("TTYPE2")

        assert isinstance(hdu, fits.BinTableHDU)
        filename = self.temp("test_table_to_hdu.fits")
        hdu.writeto(filename)

        assert fits.getval(filename, "EXTNAME", ext=1) == "MYTABLE"

    def test_masked_table_to_hdu(self):
        i = np.ma.MaskedArray([1, 2, 3], mask=[True, False, False])
        s = np.ma.MaskedArray(["a", "b", "c"], mask=[False, True, True])
        c = np.ma.MaskedArray([2.3 + 1j, 4.5 + 0j, 6.7 - 1j], mask=[True, False, True])
        f = np.ma.MaskedArray([2.3, 4.5, 6.7], mask=[True, False, True])
        table = Table([i, s, c, f], names=["i", "s", "c", "f"])
        # Check that FITS standard is used in replacing masked values.
        hdu = fits.table_to_hdu(table)
        assert isinstance(hdu, fits.BinTableHDU)
        assert hdu.header["TNULL1"] == i.fill_value
        assert_array_equal(hdu.data["i"], i.filled())
        assert_array_equal(hdu.data["s"], s.filled(""))
        assert_array_equal(hdu.data["c"], c.filled(np.nan))
        assert_array_equal(hdu.data["c"].real, c.real.filled(np.nan))
        assert_array_equal(hdu.data["c"].imag, c.imag.filled(np.nan))
        assert_array_equal(hdu.data["c"], c.filled(complex(np.nan, np.nan)))
        assert_array_equal(hdu.data["f"], f.filled(np.nan))
        filename = self.temp("test_table_to_hdu.fits")
        hdu.writeto(filename, overwrite=True)

    def test_masked_integer_arrays(self):
        # Regression test for #18817
        testfile = self.temp("test_masked_integer_arrays.fits")
        t_w = Table(
            rows=[
                [[np.ma.masked, np.ma.masked]],
                [[1, 2]],
                [[1, np.ma.masked]],
            ],
            names=["a"],
        )
        t_w.write(testfile, overwrite=True)
        t_r = Table.read(testfile)
        assert repr(t_w) == repr(t_r)

    def test_table_non_stringifyable_unit_to_hdu(self):
        table = Table(
            [[1, 2, 3], ["a", "b", "c"], [2.3, 4.5, 6.7]],
            names=["a", "b", "c"],
            dtype=["i", "U1", "f"],
        )
        table["a"].unit = u.core.IrreducibleUnit("test")

        with pytest.warns(
            AstropyUserWarning, match="The unit 'test' could not be saved"
        ) as w:
            fits.table_to_hdu(table)
        assert len(w) == 1

    def test_table_to_hdu_convert_comment_convention(self):
        """
        Regression test for https://github.com/astropy/astropy/issues/6079
        """
        table = Table(
            [[1, 2, 3], ["a", "b", "c"], [2.3, 4.5, 6.7]],
            names=["a", "b", "c"],
            dtype=["i", "U1", "f"],
        )
        table.meta["comments"] = ["This", "is", "a", "comment"]
        hdu = fits.table_to_hdu(table)

        assert hdu.header.get("comment") == ["This", "is", "a", "comment"]
        with pytest.raises(ValueError):
            hdu.header.index("comments")

    def test_table_to_hdu_filter_reserved(self):
        """
        Regression test for https://github.com/astropy/astropy/issues/9387
        """
        diag = "be ignored since it conflicts with a FITS reserved keyword"
        ins_cards = {
            "EXPTIME": 32.1,
            "XTENSION": "NEWTABLE",
            "NAXIS": 1,
            "NAXIS1": 3,
            "NAXIS2": 9,
            "PCOUNT": 42,
            "OBSERVER": "Adams",
        }
        table = Table(
            [[1, 2, 3], ["a", "b", "c"], [2.3, 4.5, 6.7]],
            names=["a", "b", "c"],
            dtype=["i4", "U1", "f8"],
        )
        table.meta.update(ins_cards)

        with pytest.warns(
            AstropyUserWarning, match=rf"Meta-data keyword \w+ will {diag}"
        ) as w:
            hdu = fits.table_to_hdu(table)

        # This relies on the warnings being raised in the order of the
        # meta dict (note that the first and last card are legitimate keys)
        assert len(w) == len(ins_cards) - 2
        for i, key in enumerate(list(ins_cards)[1:-1]):
            assert f"Meta-data keyword {key}" in str(w[i].message)

        assert hdu.header.get("XTENSION") == "BINTABLE"
        assert hdu.header.get("NAXIS") == 2
        assert hdu.header.get("NAXIS1") == 13
        assert hdu.header.get("NAXIS2") == 3
        assert hdu.header.get("PCOUNT") == 0
        np.testing.assert_almost_equal(hdu.header.get("EXPTIME"), 3.21e1)

    @pytest.mark.parametrize("card", REMOVE_KEYWORDS)
    def test_table_to_hdu_warn_reserved(self, card):
        """
        Test warning for each keyword in ..connect.REMOVE_KEYWORDS, 1 by 1
        """
        diag = "be ignored since it conflicts with a FITS reserved keyword"
        res_cards = {
            "XTENSION": "BINTABLE",
            "BITPIX": 8,
            "NAXIS": 2,
            "NAXIS1": 12,
            "NAXIS2": 3,
            "PCOUNT": 0,
            "GCOUNT": 1,
            "TFIELDS": 2,
            "THEAP": None,
        }
        ins_cards = {
            "XTENSION": "TABLE",
            "BITPIX": 16,
            "NAXIS": 1,
            "NAXIS1": 2,
            "NAXIS2": 6,
            "PCOUNT": 2,
            "GCOUNT": 2,
            "TFIELDS": 4,
            "THEAP": 36,
        }

        table = Table(
            [[1.0, 2.0, 3.0], [2.3, 4.5, 6.7]],
            names=["wavelength", "flux"],
            dtype=["f8", "f4"],
        )
        table.meta["ORIGIN"] = "Min.Silly Walks"
        table.meta[card] = ins_cards[card]
        assert table.meta.get(card) != res_cards[card]

        with pytest.warns(
            AstropyUserWarning, match=f"Meta-data keyword {card} will {diag}"
        ):
            hdu = fits.table_to_hdu(table)

        assert hdu.header.get(card) == res_cards[card]
        assert hdu.header.get("ORIGIN") == "Min.Silly Walks"

    def test_table_to_hdu_filter_incompatible(self):
        """
        Test removal of unsupported data types from header
        """
        table = Table(
            [[1, 2, 3], ["a", "b", "c"], [2.3, 4.5, 6.7]],
            names=["a", "b", "c"],
            dtype=["i4", "U1", "f8"],
        )
        table.meta.update(
            {
                "OBSDATE": "2001-05-26",
                "RAMP": np.arange(5),
                "TARGETS": {"PRIMARY": 1, "SECONDAR": 3},
            }
        )
        with pytest.warns(
            AstropyUserWarning,
            match=r"Attribute \S+ of type "
            r".+ cannot be added to FITS Header - skipping",
        ):
            hdu = fits.table_to_hdu(table)

        assert hdu.header.get("OBSDATE") == "2001-05-26"
        assert "RAMP" not in hdu.header
        assert "TARGETS" not in hdu.header

    def test_table_writeto_header(self):
        """
        Regression test for https://github.com/astropy/astropy/issues/5988
        """
        data = np.zeros((5,), dtype=[("x", float), ("y", int)])
        h_in = fits.Header()
        h_in["ANSWER"] = (42.0, "LTU&E")
        filename = self.temp("tabhdr42.fits")
        fits.writeto(filename, data=data, header=h_in, overwrite=True)
        h_out = fits.getheader(filename, ext=1)
        assert h_out["ANSWER"] == 42

    def test_image_extension_update_header(self, home_is_temp):
        """
        Test that _makehdu correctly includes the header. For example in the
        fits.update convenience function.
        """
        filename = self.temp("twoextension.fits")

        hdus = [fits.PrimaryHDU(np.zeros((10, 10))), fits.ImageHDU(np.zeros((10, 10)))]

        # Try to update a non-existent file
        with pytest.raises(FileNotFoundError, match="No such file"):
            fits.update(
                filename, np.zeros((10, 10)), header=fits.Header([("WHAT", 100)]), ext=1
            )

        fits.HDUList(hdus).writeto(filename)

        fits.update(
            filename, np.zeros((10, 10)), header=fits.Header([("WHAT", 100)]), ext=1
        )
        h_out = fits.getheader(filename, ext=1)
        assert h_out["WHAT"] == 100

    def test_printdiff(self):
        """
        Test that FITSDiff can run the different inputs without crashing.
        """

        # Testing different string input options
        assert printdiff(self.data("arange.fits"), self.data("blank.fits")) is None
        assert (
            printdiff(self.data("arange.fits"), self.data("blank.fits"), ext=0) is None
        )
        assert (
            printdiff(
                self.data("o4sp040b0_raw.fits"),
                self.data("o4sp040b0_raw.fits"),
                extname="sci",
            )
            is None
        )

        # This may seem weird, but check printdiff to see, need to test
        # incorrect second file
        with pytest.raises(OSError):
            printdiff("o4sp040b0_raw.fits", "fakefile.fits", extname="sci")

        # Test HDU object inputs
        with fits.open(self.data("stddata.fits"), mode="readonly") as in1:
            with fits.open(self.data("checksum.fits"), mode="readonly") as in2:
                assert printdiff(in1[0], in2[0]) is None

                with pytest.raises(ValueError):
                    printdiff(in1[0], in2[0], ext=0)

                assert printdiff(in1, in2) is None

                with pytest.raises(NotImplementedError):
                    printdiff(in1, in2, 0)

    def test_tabledump(self):
        """
        A simple test of the dump method.
        Also regression test for https://github.com/astropy/astropy/issues/6937
        """
        datastr = (
            '"                    1" "abc" "     3.70000007152557" "                    0"\n'
            '"                    2" "xy " "     6.69999971389771" "                    1"\n'
        )
        cdstr = (
            'c1 1J I11              ""               ""'
            '               -2147483647      ""               ""              \n'
            'c2 3A A3               ""               ""'
            '               ""               ""               ""              \n'
            'c3 1E G15.7            ""               ""'
            '               ""               3                0.4             \n'
            'c4 1L L6               ""               ""'
            '               ""               ""               ""              \n'
        )
        # copy fits file to the temp directory
        testfile = self.copy_file("tb.fits")

        # test without datafile
        fits.tabledump(testfile)
        assert os.path.isfile(self.temp("tb_1.txt"))

        # test with datafile
        fits.tabledump(testfile, datafile=self.temp("test_tb.txt"))
        assert os.path.isfile(self.temp("test_tb.txt"))

        # test with datafile and cdfile
        datafile = self.temp("data.txt")
        cdfile = self.temp("coldefs.txt")
        fits.tabledump(testfile, datafile, cdfile)
        assert os.path.isfile(datafile)
        with open(datafile) as data:
            assert data.read() == datastr
        with open(cdfile) as coldefs:
            assert coldefs.read() == cdstr

    @pytest.mark.parametrize("tablename", ["table.fits", "tb.fits"])
    def test_dump_load_round_trip(self, tablename):
        """
        A simple test of the dump/load methods; dump the data, column, and
        header files and try to reload the table from them.
        """

        # copy fits file to the temp directory
        testfile = self.copy_file(tablename)

        datafile = self.temp("data.txt")
        cdfile = self.temp("coldefs.txt")
        hfile = self.temp("header.txt")
        fits.tabledump(testfile, datafile, cdfile, hfile)

        new_tbhdu = fits.tableload(datafile, cdfile, hfile)

        with fits.open(testfile) as hdul:
            _assert_attr_col(new_tbhdu, hdul[1])

    def test_append_filename(self, home_is_temp):
        """
        Test fits.append with a filename argument.
        """
        data = np.arange(6)
        testfile = self.temp("test_append_1.fits")

        # Test case 1: creation of file
        fits.append(testfile, data=data, checksum=True)

        # Test case 2: append to existing file, with verify=True
        # Also test that additional keyword can be passed to fitsopen
        fits.append(testfile, data=data * 2, checksum=True, ignore_blank=True)

        # Test case 3: append to existing file, with verify=False
        fits.append(testfile, data=data * 3, checksum=True, verify=False)

        with fits.open(testfile, checksum=True) as hdu1:
            np.testing.assert_array_equal(hdu1[0].data, data)
            np.testing.assert_array_equal(hdu1[1].data, data * 2)
            np.testing.assert_array_equal(hdu1[2].data, data * 3)

    @pytest.mark.parametrize("mode", ["wb", "wb+", "ab", "ab+"])
    def test_append_filehandle(self, tmp_path, mode):
        """
        Test fits.append with a file handle argument.
        """
        append_file = tmp_path / "append.fits"
        with append_file.open(mode) as handle:
            fits.append(filename=handle, data=np.ones((4, 4)))

    def test_append_with_header(self):
        """
        Test fits.append with a fits Header, which triggers detection of the
        HDU class. Regression test for
        https://github.com/astropy/astropy/issues/8660
        """
        testfile = self.temp("test_append_1.fits")
        with fits.open(self.data("test0.fits")) as hdus:
            for hdu in hdus:
                fits.append(testfile, hdu.data, hdu.header, checksum=True)

        with fits.open(testfile, checksum=True) as hdus:
            assert len(hdus) == 5

    def test_pathlib(self):
        testfile = pathlib.Path(self.temp("test.fits"))
        data = np.arange(10)
        hdulist = fits.HDUList([fits.PrimaryHDU(data)])
        hdulist.writeto(testfile)

        with fits.open(testfile) as hdul:
            np.testing.assert_array_equal(hdul[0].data, data)

    def test_getdata_ext_given(self):
        prihdu = fits.PrimaryHDU(data=np.zeros((5, 5), dtype=int))
        exthdu1 = fits.ImageHDU(data=np.ones((5, 5), dtype=int))
        exthdu2 = fits.ImageHDU(data=2 * np.ones((5, 5), dtype=int))
        hdulist = fits.HDUList([prihdu, exthdu1, exthdu2])
        buf = io.BytesIO()
        hdulist.writeto(buf)

        for ext in [0, 1, 2]:
            buf.seek(0)
            data = fits.getdata(buf, ext=ext)
            assert data[0, 0] == ext

    def test_getdata_ext_given_nodata(self):
        prihdu = fits.PrimaryHDU(data=np.zeros((5, 5), dtype=int))
        exthdu1 = fits.ImageHDU(data=np.ones((5, 5), dtype=int))
        exthdu2 = fits.ImageHDU(data=None)
        hdulist = fits.HDUList([prihdu, exthdu1, exthdu2])
        buf = io.BytesIO()
        hdulist.writeto(buf)
        buf.seek(0)

        with pytest.raises(IndexError, match="No data in HDU #2."):
            fits.getdata(buf, ext=2)

    def test_getdata_ext_not_given_with_data_in_primary(self):
        prihdu = fits.PrimaryHDU(data=np.zeros((5, 5), dtype=int))
        exthdu1 = fits.ImageHDU(data=None)
        exthdu2 = fits.ImageHDU(data=None)
        hdulist = fits.HDUList([prihdu, exthdu1, exthdu2])
        buf = io.BytesIO()
        hdulist.writeto(buf)
        buf.seek(0)

        data = fits.getdata(buf)
        assert data[0, 0] == 0

    def test_getdata_ext_not_given_with_data_in_ext(self):
        # tests fallback mechanism
        prihdu = fits.PrimaryHDU(data=None)
        exthdu1 = fits.ImageHDU(data=np.ones((5, 5), dtype=int))
        exthdu2 = fits.ImageHDU(data=None)
        hdulist = fits.HDUList([prihdu, exthdu1, exthdu2])
        buf = io.BytesIO()
        hdulist.writeto(buf)
        buf.seek(0)

        data = fits.getdata(buf)
        assert data[0, 0] == 1

    def test_getdata_ext_not_given_nodata_any(self):
        # tests exception raised when there is no data in either
        # Primary HDU or first extension HDU
        prihdu = fits.PrimaryHDU(data=None)
        exthdu1 = fits.ImageHDU(data=None)
        exthdu2 = fits.ImageHDU(data=np.ones((5, 5), dtype=int))
        hdulist = fits.HDUList([prihdu, exthdu1, exthdu2])
        buf = io.BytesIO()
        hdulist.writeto(buf)
        buf.seek(0)

        with pytest.raises(
            IndexError, match="No data in either Primary or first extension HDUs."
        ):
            fits.getdata(buf)

    def test_getdata_ext_not_given_nodata_noext(self):
        # tests exception raised when there is no data in the
        # Primary HDU and there are no extension HDUs
        prihdu = fits.PrimaryHDU(data=None)
        hdulist = fits.HDUList([prihdu])
        buf = io.BytesIO()
        hdulist.writeto(buf)
        buf.seek(0)

        with pytest.raises(
            IndexError, match="No data in Primary HDU and no extension HDU found."
        ):
            fits.getdata(buf)



from astropy.timeseries.periodograms.base import *
from astropy.timeseries.periodograms.bls import *
from astropy.timeseries.periodograms.lombscargle import *
from astropy.timeseries.periodograms.lombscargle_multiband import *



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Functions for `astropy.cosmology`."""

__all__ = ["cosmology_equal", "z_at_value"]

from .comparison import cosmology_equal
from .optimize import z_at_value



# Licensed under a 3-clause BSD style license - see LICENSE.rst

__all__ = [
    "validate_non_negative",
    "validate_to_float",
    "validate_to_scalar",
    "validate_with_unit",
]

from collections.abc import Callable
from typing import TYPE_CHECKING, Any

from numpy.typing import NDArray

import astropy.units as u

if TYPE_CHECKING:
    import astropy.cosmology

FValidateCallable = Callable[[object, object, Any], Any]
_REGISTRY_FVALIDATORS: dict[str, FValidateCallable] = {}


def _register_validator(
    key: str, fvalidate: FValidateCallable | None = None
) -> FValidateCallable | Callable[[FValidateCallable], FValidateCallable]:
    """Decorator to register a new kind of validator function.

    Parameters
    ----------
    key : str
    fvalidate : callable[[object, object, Any], Any] or None, optional
        Value validation function.

    Returns
    -------
    ``validator`` or callable[``validator``]
        if validator is None returns a function that takes and registers a
        validator. This allows ``register_validator`` to be used as a
        decorator.
    """
    if key in _REGISTRY_FVALIDATORS:
        raise KeyError(f"validator {key!r} already registered with Parameter.")

    # fvalidate directly passed
    if fvalidate is not None:
        _REGISTRY_FVALIDATORS[key] = fvalidate
        return fvalidate

    # for use as a decorator
    def register(fvalidate):
        """Register validator function.

        Parameters
        ----------
        fvalidate : callable[[object, object, Any], Any]
            Validation function.

        Returns
        -------
        ``validator``
        """
        _REGISTRY_FVALIDATORS[key] = fvalidate
        return fvalidate

    return register


# ======================================================================


@_register_validator("default")
def validate_with_unit(
    cosmology: "astropy.cosmology.Cosmology",
    param: "astropy.cosmology.Parameter",
    value: Any,
) -> Any:
    """Default Parameter value validator.

    Adds/converts units if Parameter has a unit.
    """
    if param.unit is not None:
        with u.add_enabled_equivalencies(param.equivalencies):
            value = u.Quantity(value, param.unit)
    return value


@_register_validator("float")
def validate_to_float(
    cosmology: "astropy.cosmology.Cosmology",
    param: "astropy.cosmology.Parameter",
    value: Any,
) -> float:
    """Parameter value validator with units, and converted to float."""
    value = validate_with_unit(cosmology, param, value)
    return float(value)


@_register_validator("scalar")
def validate_to_scalar(
    cosmology: "astropy.cosmology.Cosmology",
    param: "astropy.cosmology.Parameter",
    value: Any,
) -> NDArray:
    """Parameter value validator where value is a scalar."""
    value = validate_with_unit(cosmology, param, value)
    if not value.isscalar:
        raise ValueError(f"{param.name} is a non-scalar quantity")
    return value


@_register_validator("non-negative")
def validate_non_negative(
    cosmology: "astropy.cosmology.Cosmology",
    param: "astropy.cosmology.Parameter",
    value: Any,
) -> float:
    """Parameter value validator where value is a positive float."""
    value = validate_to_float(cosmology, param, value)
    if value < 0.0:
        raise ValueError(f"{param.name} cannot be negative.")
    return value



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Photon component."""

__all__ = ["PhotonComponent"]

from collections.abc import Callable
from typing import Any

from numpy.typing import ArrayLike, NDArray

from astropy.cosmology._src.typing import FArray
from astropy.cosmology._src.utils import aszarr, deprecated_keywords
from astropy.units import Quantity


class PhotonComponent:
    """The cosmology has attributes and methods for the photon density."""

    Ogamma0: float
    """Omega gamma; the density/critical density of photons at z=0."""

    inv_efunc: Callable[[NDArray[Any]], NDArray[Any]]

    @deprecated_keywords("z", since="7.0")
    def Ogamma(self, z: Quantity | ArrayLike) -> FArray:
        """Return the density parameter for photons at redshift ``z``.

        Parameters
        ----------
        z : Quantity-like ['redshift'], array-like
            Input redshift.

            .. versionchanged:: 7.0
                Passing z as a keyword argument is deprecated.

        Returns
        -------
        Ogamma : array
            The energy density of photons relative to the critical density at
            each redshift.
        """
        z = aszarr(z)
        return self.Ogamma0 * (z + 1.0) ** 4 * self.inv_efunc(z) ** 2



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import erfa

from astropy.coordinates.attributes import TimeAttribute
from astropy.coordinates.baseframe import base_doc, frame_transform_graph
from astropy.coordinates.transformations import DynamicMatrixTransform
from astropy.utils.decorators import format_doc

from .baseradec import BaseRADecFrame, doc_components
from .utils import EQUINOX_J2000, get_jd12

__all__ = ["FK5"]


doc_footer = """
    Other parameters
    ----------------
    equinox : `~astropy.time.Time`
        The equinox of this frame.
"""


@format_doc(base_doc, components=doc_components, footer=doc_footer)
class FK5(BaseRADecFrame):
    """
    A coordinate or frame in the FK5 system.

    Note that this is a barycentric version of FK5 - that is, the origin for
    this frame is the Solar System Barycenter, *not* the Earth geocenter.

    The frame attributes are listed under **Other Parameters**.
    """

    equinox = TimeAttribute(default=EQUINOX_J2000, doc="The equinox time")

    @staticmethod
    def _precession_matrix(oldequinox, newequinox):
        """
        Compute and return the precession matrix for FK5 based on Capitaine et
        al. 2003/IAU2006.  Used inside some of the transformation functions.

        Parameters
        ----------
        oldequinox : `~astropy.time.Time`
            The equinox to precess from.
        newequinox : `~astropy.time.Time`
            The equinox to precess to.

        Returns
        -------
        newcoord : array
            The precession matrix to transform to the new equinox

        References
        ----------
        Hilton, J. et al., 2006, Celest.Mech.Dyn.Astron. 94, 351
        """
        # Multiply the two precession matrices (without frame bias) through J2000.0
        fromepoch_to_J2000 = erfa.bp06(*get_jd12(oldequinox, "tt"))[1].swapaxes(-2, -1)
        J2000_to_toepoch = erfa.bp06(*get_jd12(newequinox, "tt"))[1]
        return J2000_to_toepoch @ fromepoch_to_J2000


# This is the "self-transform".  Defined at module level because the decorator
#  needs a reference to the FK5 class


@frame_transform_graph.transform(DynamicMatrixTransform, FK5, FK5)
def fk5_to_fk5(fk5coord1, fk5frame2):
    return fk5coord1._precession_matrix(fk5coord1.equinox, fk5frame2.equinox)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
This package defines units that can also be used as functions of other units.
If called, their arguments are used to initialize the corresponding function
unit (e.g., ``u.mag(u.ct/u.s)``).  Note that the prefixed versions cannot be
called, as it would be unclear what, e.g., ``u.mmag(u.ct/u.s)`` would mean.

It also defines a few commonly used magnitude unit instances, like STmag,
for which the physical units are defined in `astropy.units.photometric`.

All units are also available in (and should be used through) the
`astropy.units` namespace.
"""

import unicodedata

from astropy.units import photometric
from astropy.units.core import CompositeUnit, UnitBase, _add_prefixes

from .logarithmic import DecibelUnit, DexUnit, MagUnit
from .mixin import IrreducibleFunctionUnit, RegularFunctionUnit

__all__: list[str] = []  #  Units are added at the end

_ns = globals()

###########################################################################
# Logarithmic units

# These calls are what core.def_unit would do, but we need to use the callable
# unit versions.  The actual function unit classes get added in logarithmic.

dex = IrreducibleFunctionUnit(
    ["dex"], namespace=_ns, doc="Dex: Base 10 logarithmic unit"
)
dex._function_unit_class = DexUnit

dB = RegularFunctionUnit(
    ["dB", "decibel"],
    0.1 * dex,
    namespace=_ns,
    doc="Decibel: ten per base 10 logarithmic unit",
)
dB._function_unit_class = DecibelUnit

mag = RegularFunctionUnit(
    ["mag"],
    -0.4 * dex,
    namespace=_ns,
    doc="Astronomical magnitude: -2.5 per base 10 logarithmic unit",
)
_add_prefixes(mag, namespace=_ns, prefixes=True)
mag._function_unit_class = MagUnit

STmag = mag(photometric.STflux)
STmag.__doc__ = "ST magnitude: STmag=-21.1 corresponds to 1 erg/s/cm2/A"

ABmag = mag(photometric.ABflux)
ABmag.__doc__ = "AB magnitude: ABmag=-48.6 corresponds to 1 erg/s/cm2/Hz"

M_bol = mag(photometric.Bol)
M_bol.__doc__ = (
    f"Absolute bolometric magnitude: M_bol=0 corresponds to {photometric.Bol}"
)

m_bol = mag(photometric.bol)
m_bol.__doc__ = (
    f"Apparent bolometric magnitude: m_bol=0 corresponds to {photometric.bol}"
)


###########################################################################
# DOCSTRING

__all__ += [
    n
    for n, v in _ns.items()
    if isinstance(v, (UnitBase, MagUnit)) and unicodedata.normalize("NFKC", n) == n
]

if __doc__ is not None:
    # This generates a docstring for this module that describes all of the
    # standard units defined here.
    from astropy.units.docgen import generate_unit_summary as _generate_unit_summary

    def _description(unit):
        pu = unit.physical_unit.represents
        if unit.__doc__[:2] in {"AB", "ST"}:
            pu = 1.0 * CompositeUnit(1.0, pu.bases, pu.powers)
        return "".join(
            unit.__doc__.partition("corresponds to ")[:-1]
            + (f":math:`{pu.to_string(format='latex')[1:-1]}`",)
        )

    template = """
   * - ``{}``
     - {}
     - {}
"""

    __doc__ += (
        _generate_unit_summary(globals())
        + """
.. list-table:: Available Magnitude Units
   :header-rows: 1
   :widths: 10 50 10

   * - Unit
     - Description
     - Represents
"""
        + "".join(
            [
                template.format(key, _description(val), val)
                for key, val in globals().items()
                if isinstance(val, MagUnit)
            ]
        )
    )



import numpy as np
import pytest
from numpy.testing import assert_equal

from astropy.coordinates.polarization import (
    StokesCoord,
    StokesSymbol,
    custom_stokes_symbol_mapping,
)
from astropy.table import Table, vstack
from astropy.utils import unbroadcast


def test_scalar():
    sk = StokesCoord(2)
    assert repr(sk) == "StokesCoord('Q')" == str(sk)
    assert sk.value == 2.0
    assert sk.symbol == "Q"


def test_vector():
    # This also checks that floats are rounded when converting
    # to strings
    values = [1.2, 1.8, 2.0, 2.2, 2.8]
    sk = StokesCoord(values)
    assert_equal(sk.value, values)
    assert_equal(sk.symbol, np.array(["I", "Q", "Q", "Q", "U"]))
    assert repr(sk) == "StokesCoord(['I', 'Q', 'Q', 'Q', 'U'])" == str(sk)


def test_vector_list_init():
    sk = StokesCoord(["I", "Q", "Q", "U", "U"])
    assert repr(sk) == "StokesCoord(['I', 'Q', 'Q', 'U', 'U'])" == str(sk)
    assert_equal(sk.symbol, np.array(["I", "Q", "Q", "U", "U"]))


def test_undefined():
    sk = StokesCoord(np.arange(-10, 7))
    assert_equal(sk.symbol, "? ? YX XY YY XX LR RL LL RR ? I Q U V ? ?".split())  # noqa: SIM905


def test_undefined_init():
    with pytest.raises(Exception, match="Unknown stokes symbols.*Spam"):
        StokesCoord("Spam")


def test_custom_symbol_mapping():
    custom_mapping = {
        10000: StokesSymbol("A"),
        10001: StokesSymbol("B"),
        10002: StokesSymbol("C"),
        10003: StokesSymbol("D"),
    }

    # Check that we can supply a custom mapping
    with custom_stokes_symbol_mapping(custom_mapping):
        values = [0.6, 1.7, 10000.1, 10002.4]
        sk1 = StokesCoord(values)
        assert repr(sk1) == "StokesCoord(['I', 'Q', 'A', 'C'])" == str(sk1)
        assert_equal(sk1.value, values)
        assert_equal(sk1.symbol, np.array(["I", "Q", "A", "C"]))

    # Check that the mapping is not active outside the context manager
    assert_equal(sk1.symbol, np.array(["I", "Q", "?", "?"]))

    # Also not for new StokesCoords
    sk2 = StokesCoord(values)
    assert_equal(sk2.symbol, np.array(["I", "Q", "?", "?"]))


def test_custom_symbol_mapping_overlap():
    # Make a custom mapping that overlaps with some of the existing values

    custom_mapping = {
        3: StokesSymbol("A"),
        4: StokesSymbol("B"),
        5: StokesSymbol("C"),
        6: StokesSymbol("D"),
    }

    with custom_stokes_symbol_mapping(custom_mapping):
        sk = StokesCoord(np.arange(1, 7))
        assert_equal(sk.symbol, np.array(["I", "Q", "A", "B", "C", "D"]))


def test_custom_symbol_mapping_replace():
    # Check that we can replace the mapping completely

    custom_mapping = {
        3: StokesSymbol("A"),
        4: StokesSymbol("B"),
        5: StokesSymbol("C"),
        6: StokesSymbol("D"),
    }

    with custom_stokes_symbol_mapping(custom_mapping, replace=True):
        sk = StokesCoord(np.arange(1, 7))
        assert_equal(sk.symbol, np.array(["?", "?", "A", "B", "C", "D"]))


def test_comparison_scalar():
    sk = StokesCoord(np.arange(1, 6))
    assert_equal("Q" == sk, [False, True, False, False, False])
    assert_equal(sk == 1, [True, False, False, False, False])
    assert_equal(sk == "Q", [False, True, False, False, False])


def test_comparison_vector():
    sk = StokesCoord(np.arange(1, 6))
    assert_equal(
        sk == np.array(["I", "Q", "I", "I", "Q"]), [True, True, False, False, False]
    )


def test_comparison_other_coord():
    sk1 = StokesCoord(np.arange(1, 6))
    sk2 = StokesCoord("I")
    assert_equal(sk1 == sk2, [True, False, False, False, False])
    sk3 = StokesCoord(np.repeat(2, 5))
    assert_equal(sk1 == sk3, [False, True, False, False, False])


def test_efficient():
    # Make sure that if we pass a broadcasted array in we get a broadcasted
    # array of symbols.

    values = np.broadcast_to(np.arange(1, 5, dtype=float), (512, 256, 4))
    sk = StokesCoord(values, copy=False)

    assert sk.symbol.shape == (512, 256, 4)
    assert unbroadcast(sk.value).shape == (4,)
    assert unbroadcast(sk.symbol).shape == (4,)
    assert_equal(unbroadcast(sk.symbol), np.array(["I", "Q", "U", "V"]))


def test_broadcast_to():
    sk = StokesCoord(np.arange(1, 5, dtype=int), copy=False)
    sk2 = np.broadcast_to(sk, (512, 256, 4))

    assert sk2.symbol.shape == (512, 256, 4)
    assert unbroadcast(sk2.value).shape == (4,)
    assert unbroadcast(sk2.symbol).shape == (4,)
    assert_equal(unbroadcast(sk.symbol), np.array(["I", "Q", "U", "V"]))


def test_table_vstack_stokes():
    sk = StokesCoord(np.arange(1, 5, dtype=int), copy=False)
    tt = Table([sk])
    assert isinstance(tt["col0"], StokesCoord)
    assert np.allclose(tt["col0"].value, np.arange(1, 5, dtype=int))

    sk2 = StokesCoord([1, 2, 2, 2, 4, 5])
    tt2 = Table([sk2])
    assert isinstance(tt2["col0"], StokesCoord)
    assert np.allclose(tt2["col0"].value, sk2.value)

    tt3 = vstack([tt, tt2])
    assert isinstance(tt3["col0"], StokesCoord)
    assert len(tt3) == 10
    assert np.allclose(tt3["col0"].value, np.array([1, 2, 3, 4, 1, 2, 2, 2, 4, 5]))


def test_init_copy():
    input = np.arange(1, 5, dtype=int)
    sk1 = StokesCoord(input, copy=False)
    assert sk1._data is input

    skc = StokesCoord(input, copy=True)
    assert skc._data is not input

    sk2 = StokesCoord(sk1)
    assert sk1._data is sk2._data

    sk3 = StokesCoord(sk1, copy=True)
    assert sk1._data is not sk3._data
    assert np.allclose(sk1._data, sk3._data)


def test_init_error():
    with pytest.raises(ValueError, match="object array"):
        StokesCoord(None)



# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""Utilities for generating type stubs for unit definition modules.

This module is meant for internal use only (in setup.py).
"""

__all__: list[str] = []

import inspect
from collections.abc import Iterable
from pathlib import Path
from types import FunctionType, ModuleType
from typing import Final

from astropy import units as u
from astropy.units import cds

UNIT_DEFINITION_MODULES: Final = (
    u.astrophys,
    cds,
    u.cgs,
    u.imperial,
    u.misc,
    u.photometric,
    u.required_by_vounit,
    u.si,
    u.function.units,
)


def generate_stub(output_dir: Path, module: ModuleType) -> None:
    members = {}
    functions = {}
    for name in module.__all__:
        member = getattr(module, name)
        if isinstance(member, FunctionType):
            # Neither of the functions in unit definition modules have any parameters.
            functions[name] = inspect.get_annotations(member)["return"]
        else:
            members[name] = type(member)
    import_lines = sorted(
        f"from {cls.__module__} import {cls.__name__}\n"
        for cls in set(members.values()).union(functions.values())
    )
    with (
        output_dir.joinpath(*module.__name__.split("."))
        .with_suffix(".pyi")
        .open("wt", encoding="utf-8") as stub_file
    ):
        stub_file.write("# This stub file was automatically generated.\n\n")

        stub_file.write("__all__ = [\n")
        stub_file.writelines(f'    "{name}",\n' for name in module.__all__)
        stub_file.write("]\n\n")

        stub_file.writelines(sorted(import_lines))
        stub_file.write("\n")

        stub_file.writelines(
            f"{member}: {cls.__name__}\n" for member, cls in members.items()
        )

        stub_file.writelines(
            f"\ndef {name}() -> {return_type.__name__}: ...\n"
            for name, return_type in functions.items()
        )


def create_stubs(
    output_dir: Path, modules: Iterable[ModuleType] = UNIT_DEFINITION_MODULES
) -> None:
    for module in modules:
        generate_stub(output_dir, module)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import pickle

import pytest

from astropy.io.misc import fnpickle, fnunpickle
from astropy.utils.exceptions import AstropyDeprecationWarning


def test_fnpickling_simple(tmp_path):
    """
    Tests the `fnpickle` and `fnupickle` functions' basic operation by
    pickling and unpickling a string, using both a filename and a
    file.
    """
    fn = str(tmp_path / "test1.pickle")

    obj1 = "astring"
    with pytest.warns(
        AstropyDeprecationWarning, match="Use pickle from standard library"
    ):
        fnpickle(obj1, fn)
        res = fnunpickle(fn, 0)
        assert obj1 == res

        # now try with a file-like object instead of a string
        with open(fn, "wb") as f:
            fnpickle(obj1, f)
        with open(fn, "rb") as f:
            res = fnunpickle(f)
            assert obj1 == res


class ToBePickled:
    def __init__(self, item):
        self.item = item

    def __eq__(self, other):
        if isinstance(other, ToBePickled):
            return self.item == other.item
        else:
            return False


def test_fnpickling_class(tmp_path):
    """
    Tests the `fnpickle` and `fnupickle` functions' ability to pickle
    and unpickle custom classes.
    """
    fn = str(tmp_path / "test2.pickle")

    obj1 = "astring"
    obj2 = ToBePickled(obj1)
    with pytest.warns(
        AstropyDeprecationWarning, match="Use pickle from standard library"
    ):
        fnpickle(obj2, fn)
        res = fnunpickle(fn)
    assert res == obj2


def test_fnpickling_protocol(tmp_path):
    """
    Tests the `fnpickle` and `fnupickle` functions' ability to pickle
    and unpickle pickle files from all protocols.
    """
    obj1 = "astring"
    obj2 = ToBePickled(obj1)

    for p in range(pickle.HIGHEST_PROTOCOL + 1):
        fn = str(tmp_path / f"testp{p}.pickle")
        with pytest.warns(
            AstropyDeprecationWarning, match="Use pickle from standard library"
        ):
            fnpickle(obj2, fn, protocol=p)
            res = fnunpickle(fn)
        assert res == obj2


def test_fnpickling_many(tmp_path):
    """
    Tests the `fnpickle` and `fnupickle` functions' ability to pickle
    and unpickle multiple objects from a single file.
    """

    fn = str(tmp_path / "test3.pickle")

    # now try multiples
    obj3 = 328.3432
    obj4 = "blahblahfoo"
    with pytest.warns(
        AstropyDeprecationWarning, match="Use pickle from standard library"
    ):
        fnpickle(obj3, fn)
        fnpickle(obj4, fn, append=True)

        res = fnunpickle(fn, number=-1)
        assert len(res) == 2
        assert res[0] == obj3
        assert res[1] == obj4

        fnpickle(obj4, fn, append=True)
        res = fnunpickle(fn, number=2)
        assert len(res) == 2

        with pytest.raises(EOFError):
            fnunpickle(fn, number=5)



r"""
A helper script to automate system info in bug reports.

This can be run as::

    python -c "import astropy; astropy.system_info()"

(note that the interpreter might be called python3 instead of python, depending
on your system's details)

This script should not require anything else than the standard library and is
subject to change without notice.
"""

__all__ = ["system_info"]

import platform
from importlib.metadata import version
from importlib.util import find_spec


def _header(name: str) -> list[str]:
    return [name, "-" * len(name)]


def _report_platform() -> list[str]:
    return [
        f"{platform.platform() = }",
        f"{platform.version() = }",
        f"{platform.python_version() = }",
    ]


def _report_packages() -> list[str]:
    pkgs = [
        "astropy",
        "numpy",
        "scipy",
        "matplotlib",
        "pandas",
    ]
    lines = [f"{p:<20} {'--' if find_spec(p) is None else version(p)}" for p in pkgs]

    # pyerfa may not export its package metadata (as of 2.0.1.4)
    try:
        import erfa
    except ImportError:
        v = "--"
    else:
        v = erfa.__version__
    p = "pyerfa"
    lines.append(f"{p:<20} {v}")
    return lines


def system_info() -> None:
    """Print relevant system information for astropy bug reports.

    Examples
    --------
    >>> import astropy
    >>> astropy.system_info()  # doctest: +ELLIPSIS
    platform
    --------
    platform.platform() = ...
    platform.version() = ...
    platform.python_version() = ...
    <BLANKLINE>
    packages
    --------
    astropy              ...
    numpy                ...
    scipy                ...
    matplotlib           ...
    pandas               ...
    pyerfa               ...

    """
    report_lines = (
        *_header("platform"),
        *_report_platform(),
        "",
        *_header("packages"),
        *_report_packages(),
    )
    print("\n".join(report_lines))



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import numpy as np
import pytest
from numpy.testing import assert_allclose

from astropy.utils import NumpyRNGContext
from astropy.utils.masked import Masked
from astropy.visualization.interval import (
    AsymmetricPercentileInterval,
    ManualInterval,
    MinMaxInterval,
    PercentileInterval,
    SymmetricInterval,
    ZScaleInterval,
)


class TestInterval:
    data = np.linspace(-20.0, 60.0, 100)

    def test_manual(self):
        interval = ManualInterval(-10.0, +15.0)
        vmin, vmax = interval.get_limits(self.data)
        assert_allclose(vmin, -10.0)
        assert_allclose(vmax, +15.0)

    def test_manual_defaults(self):
        interval = ManualInterval(vmin=-10.0)
        vmin, vmax = interval.get_limits(self.data)
        assert_allclose(vmin, -10.0)
        assert_allclose(vmax, np.max(self.data))

        interval = ManualInterval(vmax=15.0)
        vmin, vmax = interval.get_limits(self.data)
        assert_allclose(vmin, np.min(self.data))
        assert_allclose(vmax, 15.0)

    def test_manual_zero_limit(self):
        # Regression test for a bug that caused ManualInterval to compute the
        # limit (min or max) if it was set to zero.
        interval = ManualInterval(vmin=0, vmax=0)
        vmin, vmax = interval.get_limits(self.data)
        assert_allclose(vmin, 0)
        assert_allclose(vmax, 0)

    def test_manual_defaults_with_nan(self):
        interval = ManualInterval()
        data = np.copy(self.data)
        data[0] = np.nan
        vmin, vmax = interval.get_limits(self.data)
        assert_allclose(vmin, -20)
        assert_allclose(vmax, +60)

    def test_minmax(self):
        interval = MinMaxInterval()
        vmin, vmax = interval.get_limits(self.data)
        assert_allclose(vmin, -20.0)
        assert_allclose(vmax, +60.0)

    def test_percentile(self):
        interval = PercentileInterval(62.2)
        vmin, vmax = interval.get_limits(self.data)
        assert_allclose(vmin, -4.88)
        assert_allclose(vmax, 44.88)

    def test_asymmetric_percentile(self):
        interval = AsymmetricPercentileInterval(10.5, 70.5)
        vmin, vmax = interval.get_limits(self.data)
        assert_allclose(vmin, -11.6)
        assert_allclose(vmax, 36.4)

    def test_asymmetric_percentile_nsamples(self):
        with NumpyRNGContext(12345):
            interval = AsymmetricPercentileInterval(10.5, 70.5, n_samples=20)
            vmin, vmax = interval.get_limits(self.data)
        assert_allclose(vmin, -14.367676767676768)
        assert_allclose(vmax, 40.266666666666666)

    def test_symmetric_interval_manual(self):
        interval = SymmetricInterval(radius=40)
        vmin, vmax = interval.get_limits(self.data)
        assert_allclose(vmin, -40.0)
        assert_allclose(vmax, +40.0)

        interval = SymmetricInterval(radius=100, midpoint=10)
        vmin, vmax = interval.get_limits(self.data)
        assert_allclose(vmin, -90.0)
        assert_allclose(vmax, +110.0)

    def test_symmetric_interval_auto(self):
        interval = SymmetricInterval()
        vmin, vmax = interval.get_limits(self.data)
        assert_allclose(vmin, -60.0)
        assert_allclose(vmax, +60.0)

        interval = SymmetricInterval(midpoint=50)
        vmin, vmax = interval.get_limits(self.data)
        assert_allclose(vmin, -20.0)
        assert_allclose(vmax, +120.0)


class TestIntervalList(TestInterval):
    # Make sure intervals work with lists
    data = np.linspace(-20.0, 60.0, 100).tolist()


class TestInterval2D(TestInterval):
    # Make sure intervals work with 2d arrays
    data = np.linspace(-20.0, 60.0, 100).reshape(100, 1)


class TestIntervalMaskedArray(TestInterval):
    # Make sure intervals work with MaskedArray
    data = np.concatenate((np.linspace(-20.0, 60.0, 100), np.full(100, 1e6)))
    data = np.ma.MaskedArray(data, data > 1000)


class TestIntervalMaskedNDArray(TestInterval):
    # Make sure intervals work with MaskedArray
    data = np.concatenate((np.linspace(-20.0, 60.0, 100), np.full(100, 1e6)))
    data = Masked(data, data > 1000)


def test_zscale():
    np.random.seed(42)
    data = np.random.randn(100, 100) * 5 + 10
    interval = ZScaleInterval()
    vmin, vmax = interval.get_limits(data)
    assert_allclose(vmin, -9.6, atol=0.1)
    assert_allclose(vmax, 25.4, atol=0.1)

    data = list(range(1000)) + [np.nan]
    interval = ZScaleInterval()
    vmin, vmax = interval.get_limits(data)
    assert_allclose(vmin, 0, atol=0.1)
    assert_allclose(vmax, 999, atol=0.1)

    data = list(range(100))
    interval = ZScaleInterval()
    vmin, vmax = interval.get_limits(data)
    assert_allclose(vmin, 0, atol=0.1)
    assert_allclose(vmax, 99, atol=0.1)


def test_zscale_npoints():
    """
    Regression test to ensure ZScaleInterval returns the minimum and
    maximum of the data if the number of data points is less than
    ``min_pixels``.
    """

    data = np.arange(4).reshape((2, 2))
    interval = ZScaleInterval(min_npixels=5)
    vmin, vmax = interval.get_limits(data)
    assert vmin == 0
    assert vmax == 3


def test_integers():
    # Need to make sure integers get cast to float
    interval = MinMaxInterval()
    values = interval([1, 3, 4, 5, 6])
    assert_allclose(values, [0.0, 0.4, 0.6, 0.8, 1.0])

    # Don't accept integer array in output
    out = np.zeros(5, dtype=int)
    with pytest.raises(
        TypeError, match=r"Can only do in-place scaling for floating-point arrays"
    ):
        values = interval([1, 3, 4, 5, 6], out=out)

    # But integer input and floating point output is fine
    out = np.zeros(5, dtype=float)
    interval([1, 3, 4, 5, 6], out=out)
    assert_allclose(out, [0.0, 0.4, 0.6, 0.8, 1.0])


def test_constant_data():
    """Test intervals with constant data (avoiding divide-by-zero)."""
    shape = (10, 10)
    data = np.ones(shape)
    interval = MinMaxInterval()
    limits = interval.get_limits(data)
    values = interval(data)
    assert_allclose(limits, (1.0, 1.0))
    assert_allclose(values, np.zeros(shape))



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""|Cosmology| <-> |Table| I/O, using |Cosmology.to_format| and |Cosmology.from_format|.

This module provides functions to transform a |Cosmology| object to and from a |Table|
object. The functions are registered with ``convert_registry`` under the format name
"astropy.table". |Table| itself has an abundance of I/O methods, making this conversion
useful for further interoperability with other formats.

A Cosmology as a `~astropy.table.QTable` will have the cosmology's name and parameters
as columns.

    >>> from astropy.cosmology import Planck18
    >>> ct = Planck18.to_format("astropy.table")
    >>> ct
    <QTable length=1>
        name        H0        Om0    Tcmb0    Neff      m_nu      Ob0
                km / (Mpc s)            K                 eV
        str8     float64    float64 float64 float64  float64[3] float64
    -------- ------------ ------- ------- ------- ----------- -------
    Planck18        67.66 0.30966  2.7255   3.046 0.0 .. 0.06 0.04897

The cosmological class and other metadata, e.g. a paper reference, are in the Table's
metadata.

    >>> ct.meta
    {..., 'cosmology': 'FlatLambdaCDM'}


Cosmology supports the astropy Table-like protocol (see :ref:`Table-like Objects`) to
the same effect:

.. code-block::

    >>> from astropy.table import QTable
    >>> QTable(Planck18)
    <QTable length=1>
      name        H0        Om0    Tcmb0    Neff      m_nu      Ob0
             km / (Mpc s)            K                 eV
      str8     float64    float64 float64 float64  float64[3] float64
    -------- ------------ ------- ------- ------- ----------- -------
    Planck18        67.66 0.30966  2.7255   3.046 0.0 .. 0.06 0.04897


To move the cosmology class from the metadata to a Table row, set the
``cosmology_in_meta`` argument to `False`:

    >>> Planck18.to_format("astropy.table", cosmology_in_meta=False)
    <QTable length=1>
        cosmology     name        H0        Om0    Tcmb0    Neff      m_nu      Ob0
                            km / (Mpc s)            K                 eV
        str13       str8     float64    float64 float64 float64  float64[3] float64
    ------------- -------- ------------ ------- ------- ------- ----------- -------
    FlatLambdaCDM Planck18        67.66 0.30966  2.7255   3.046 0.0 .. 0.06 0.04897

Astropy recommends `~astropy.table.QTable` for tables with |Quantity| columns. However
the returned type may be overridden using the ``cls`` argument:

    >>> from astropy.table import Table
    >>> Planck18.to_format("astropy.table", cls=Table)
    <Table length=1>
    ...

Fields of the cosmology may be renamed using the ``rename`` argument.

    >>> Planck18.to_format("astropy.table", rename={"H0": "Hubble"})
    <QTable length=1>
        name      Hubble      Om0    Tcmb0    Neff      m_nu      Ob0
                km / (Mpc s)            K                 eV
        str8     float64    float64 float64 float64  float64[3] float64
    -------- ------------ ------- ------- ------- ----------- -------
    Planck18        67.66 0.30966  2.7255   3.046 0.0 .. 0.06 0.04897

Appropriately formatted tables can be converted to |Cosmology| instances. Since the
|Table| can hold arbitrary metadata, we can faithfully round-trip a |Cosmology| through
|Table|, e.g. to construct a ``Planck18`` cosmology identical to the instance from which
it was generated.

    >>> ct = Planck18.to_format("astropy.table")

    >>> cosmo = Cosmology.from_format(ct, format="astropy.table")
    >>> cosmo
    FlatLambdaCDM(name='Planck18', H0=<Quantity 67.66 km / (Mpc s)>, Om0=0.30966, Tcmb0=<Quantity 2.7255 K>, Neff=3.046, m_nu=<Quantity [0.  , 0.  , 0.06] eV>, Ob0=0.04897)
    >>> cosmo == Planck18
    True

The ``cosmology`` information (row or metadata) may be omitted if the cosmology class
(or its string name) is passed as the ``cosmology`` keyword argument to
|Cosmology.from_format|.

    >>> del ct.meta["cosmology"]  # remove cosmology from metadata
    >>> Cosmology.from_format(ct, cosmology="FlatLambdaCDM")
    FlatLambdaCDM(name='Planck18', H0=<Quantity 67.66 km / (Mpc s)>, Om0=0.30966, Tcmb0=<Quantity 2.7255 K>, Neff=3.046, m_nu=<Quantity [0.  , 0.  , 0.06] eV>, Ob0=0.04897)

Alternatively, specific cosmology classes can be used to parse the data.

    >>> from astropy.cosmology import FlatLambdaCDM
    >>> FlatLambdaCDM.from_format(ct)
     FlatLambdaCDM(name='Planck18', H0=<Quantity 67.66 km / (Mpc s)>, Om0=0.30966, Tcmb0=<Quantity 2.7255 K>, Neff=3.046, m_nu=<Quantity [0.  , 0.  , 0.06] eV>, Ob0=0.04897)

When using a specific cosmology class, the class' default parameter values are used to
fill in any missing information.

    >>> del ct["Tcmb0"]  # show FlatLambdaCDM provides default
    >>> FlatLambdaCDM.from_format(ct)
     FlatLambdaCDM(name='Planck18', H0=<Quantity 67.66 km / (Mpc s)>, Om0=0.30966, Tcmb0=<Quantity 0. K>, Neff=3.046, m_nu=None, Ob0=0.04897)

For tables with multiple rows of cosmological parameters, the ``index`` argument is
needed to select the correct row. The index can be an integer for the row number or, if
the table is indexed by a column, the value of that column. If the table is not indexed
and ``index`` is a string, the "name" column is used as the indexing column.

Here is an example where ``index`` is needed and can be either an integer (for the row
number) or the name of one of the cosmologies, e.g. 'Planck15'.

    >>> from astropy.cosmology import Planck13, Planck15, Planck18
    >>> from astropy.table import vstack
    >>> cts = vstack([c.to_format("astropy.table")
    ...               for c in (Planck13, Planck15, Planck18)],
    ...              metadata_conflicts='silent')
    >>> cts
    <QTable length=3>
        name        H0        Om0    Tcmb0    Neff      m_nu      Ob0
                km / (Mpc s)            K                 eV
        str8     float64    float64 float64 float64  float64[3]  float64
    -------- ------------ ------- ------- ------- ----------- --------
    Planck13        67.77 0.30712  2.7255   3.046 0.0 .. 0.06 0.048252
    Planck15        67.74  0.3075  2.7255   3.046 0.0 .. 0.06   0.0486
    Planck18        67.66 0.30966  2.7255   3.046 0.0 .. 0.06  0.04897

    >>> cosmo = Cosmology.from_format(cts, index=1, format="astropy.table")
    >>> cosmo == Planck15
    True

Fields in the table can be renamed to match the `~astropy.cosmology.Cosmology` class'
signature using the ``rename`` argument. This is useful when the table's column names do
not match the class' parameter names.

    >>> renamed_table = Planck18.to_format("astropy.table", rename={"H0": "Hubble"})
    >>> renamed_table
    <QTable length=1>
        name      Hubble      Om0    Tcmb0    Neff      m_nu      Ob0
                km / (Mpc s)            K                 eV
        str8     float64    float64 float64 float64  float64[3] float64
    -------- ------------ ------- ------- ------- ----------- -------
    Planck18        67.66 0.30966  2.7255   3.046 0.0 .. 0.06 0.04897

    >>> cosmo = Cosmology.from_format(renamed_table, format="astropy.table",
    ...                               rename={"Hubble": "H0"})
    >>> cosmo == Planck18
    True
"""

from collections.abc import Mapping
from typing import TypeVar

import numpy as np

from astropy.cosmology._src.core import Cosmology
from astropy.cosmology._src.io.connect import convert_registry
from astropy.cosmology._src.typing import _CosmoT
from astropy.table import Column, QTable, Table

from .mapping import to_mapping
from .row import from_row
from .utils import convert_parameter_to_column

_TableT = TypeVar("_TableT", bound=Table)


def from_table(
    table: Table,
    index: int | str | None = None,
    *,
    move_to_meta: bool = False,
    cosmology: str | type[_CosmoT] | None = None,
    rename: Mapping[str, str] | None = None,
) -> _CosmoT:
    """Instantiate a `~astropy.cosmology.Cosmology` from a |QTable|.

    Parameters
    ----------
    table : `~astropy.table.Table`
        The object to parse into a |Cosmology|.
    index : int, str, or None, optional
        Needed to select the row in tables with multiple rows. ``index`` can be an
        integer for the row number or, if the table is indexed by a column, the value of
        that column. If the table is not indexed and ``index`` is a string, the "name"
        column is used as the indexing column.

    move_to_meta : bool (optional, keyword-only)
        Whether to move keyword arguments that are not in the Cosmology class' signature
        to the Cosmology's metadata. This will only be applied if the Cosmology does NOT
        have a keyword-only argument (e.g. ``**kwargs``). Arguments moved to the
        metadata will be merged with existing metadata, preferring specified metadata in
        the case of a merge conflict (e.g. for ``Cosmology(meta={'key':10}, key=42)``,
        the ``Cosmology.meta`` will be ``{'key': 10}``).

    cosmology : str or type or None (optional, keyword-only)
        The cosmology class (or string name thereof) to use when constructing the
        cosmology instance. The class also provides default parameter values, filling in
        any non-mandatory arguments missing in 'table'.

    rename : Mapping[str, str] or None (optional, keyword-only)
        A mapping of column names in 'table' to field names of the |Cosmology| class.

    Returns
    -------
    `~astropy.cosmology.Cosmology`

    Examples
    --------
    To see loading a `~astropy.cosmology.Cosmology` from a Table with ``from_table``, we
    will first make a |QTable| using :func:`~astropy.cosmology.Cosmology.to_format`.

        >>> from astropy.cosmology import Cosmology, Planck18
        >>> ct = Planck18.to_format("astropy.table")
        >>> ct
        <QTable length=1>
          name        H0        Om0    Tcmb0    Neff      m_nu      Ob0
                 km / (Mpc s)            K                 eV
          str8     float64    float64 float64 float64  float64[3] float64
        -------- ------------ ------- ------- ------- ----------- -------
        Planck18        67.66 0.30966  2.7255   3.046 0.0 .. 0.06 0.04897

    Now this table can be used to load a new cosmological instance identical to the
    ``Planck18`` cosmology from which it was generated.

        >>> cosmo = Cosmology.from_format(ct, format="astropy.table")
        >>> cosmo
        FlatLambdaCDM(name='Planck18', H0=<Quantity 67.66 km / (Mpc s)>, Om0=0.30966, Tcmb0=<Quantity 2.7255 K>, Neff=3.046, m_nu=<Quantity [0.  , 0.  , 0.06] eV>, Ob0=0.04897)

    The ``cosmology`` information (column or metadata) may be omitted if the cosmology
    class (or its string name) is passed as the ``cosmology`` keyword argument to
    |Cosmology.from_format|.

        >>> del ct.meta["cosmology"]  # remove cosmology from metadata
        >>> Cosmology.from_format(ct, cosmology="FlatLambdaCDM")
        FlatLambdaCDM(name='Planck18', H0=<Quantity 67.66 km / (Mpc s)>, Om0=0.30966, Tcmb0=<Quantity 2.7255 K>, Neff=3.046, m_nu=<Quantity [0.  , 0.  , 0.06] eV>, Ob0=0.04897)

    Alternatively, specific cosmology classes can be used to parse the data.

        >>> from astropy.cosmology import FlatLambdaCDM
        >>> FlatLambdaCDM.from_format(ct)
         FlatLambdaCDM(name='Planck18', H0=<Quantity 67.66 km / (Mpc s)>, Om0=0.30966, Tcmb0=<Quantity 2.7255 K>, Neff=3.046, m_nu=<Quantity [0.  , 0.  , 0.06] eV>, Ob0=0.04897)

    When using a specific cosmology class, the class' default parameter values are used
    to fill in any missing information.

        >>> del ct["Tcmb0"]  # show FlatLambdaCDM provides default
        >>> FlatLambdaCDM.from_format(ct)
         FlatLambdaCDM(name='Planck18', H0=<Quantity 67.66 km / (Mpc s)>, Om0=0.30966, Tcmb0=<Quantity 0. K>, Neff=3.046, m_nu=None, Ob0=0.04897)

    For tables with multiple rows of cosmological parameters, the ``index`` argument is
    needed to select the correct row. The index can be an integer for the row number or,
    if the table is indexed by a column, the value of that column. If the table is not
    indexed and ``index`` is a string, the "name" column is used as the indexing column.

    Here is an example where ``index`` is needed and can be either an integer (for the
    row number) or the name of one of the cosmologies, e.g. 'Planck15'.

        >>> from astropy.cosmology import Planck13, Planck15, Planck18
        >>> from astropy.table import vstack
        >>> cts = vstack([c.to_format("astropy.table")
        ...               for c in (Planck13, Planck15, Planck18)],
        ...              metadata_conflicts='silent')
        >>> cts
        <QTable length=3>
          name        H0        Om0    Tcmb0    Neff      m_nu      Ob0
                 km / (Mpc s)            K                 eV
          str8     float64    float64 float64 float64  float64[3]  float64
        -------- ------------ ------- ------- ------- ----------- --------
        Planck13        67.77 0.30712  2.7255   3.046 0.0 .. 0.06 0.048252
        Planck15        67.74  0.3075  2.7255   3.046 0.0 .. 0.06   0.0486
        Planck18        67.66 0.30966  2.7255   3.046 0.0 .. 0.06  0.04897

        >>> cosmo = Cosmology.from_format(cts, index="Planck15", format="astropy.table")
        >>> cosmo == Planck15
        True

    Fields in the table can be renamed to match the `~astropy.cosmology.Cosmology`
    class' signature using the ``rename`` argument. This is useful when the table's
    column names do not match the class' parameter names.

        >>> renamed_table = Planck18.to_format("astropy.table", rename={"H0": "Hubble"})
        >>> renamed_table
        <QTable length=1>
          name      Hubble      Om0    Tcmb0    Neff      m_nu      Ob0
                 km / (Mpc s)            K                 eV
          str8     float64    float64 float64 float64  float64[3] float64
        -------- ------------ ------- ------- ------- ----------- -------
        Planck18        67.66 0.30966  2.7255   3.046 0.0 .. 0.06 0.04897

        >>> cosmo = Cosmology.from_format(renamed_table, format="astropy.table",
        ...                               rename={"Hubble": "H0"})
        >>> cosmo == Planck18
        True

    For further examples, see :doc:`astropy:cosmology/io`.
    """
    # Get row from table
    # string index uses the indexed column on the table to find the row index.
    if isinstance(index, str):
        if not table.indices:  # no indexing column, find by string match
            nc = "name"  # default name column
            if rename is not None:  # from inverted `rename`
                for key, value in rename.items():
                    if value == "name":
                        nc = key
                        break

            indices = np.where(table[nc] == index)[0]
        else:  # has indexing column
            indices = table.loc_indices[index]  # need to convert to row index (int)

        if isinstance(indices, (int, np.integer)):  # loc_indices
            index = indices
        elif len(indices) == 1:  # only happens w/ np.where
            index = indices[0]
        elif len(indices) == 0:  # matches from loc_indices
            raise KeyError(f"No matches found for key {indices}")
        else:  # like the Highlander, there can be only 1 Cosmology
            raise ValueError(f"more than one cosmology found for key {indices}")

    # no index is needed for a 1-row table. For a multi-row table...
    if index is None:
        if len(table) != 1:  # multi-row table and no index
            raise ValueError(
                "need to select a specific row (e.g. index=1) when "
                "constructing a Cosmology from a multi-row table."
            )
        else:  # single-row table
            index = 0
    row = table[index]  # index is now the row index (int)

    # parse row to cosmo
    return from_row(row, move_to_meta=move_to_meta, cosmology=cosmology, rename=rename)


def to_table(
    cosmology: Cosmology,
    *args: object,
    cls: type[_TableT] = QTable,
    cosmology_in_meta: bool = True,
    rename: Mapping[str, str] | None = None,
) -> _TableT:
    """Serialize the cosmology into a `~astropy.table.QTable`.

    Parameters
    ----------
    cosmology : `~astropy.cosmology.Cosmology`
        The cosmology instance to convert to a table.
    *args : object
        Not used. Needed for compatibility with
        `~astropy.io.registry.UnifiedReadWriteMethod`
    cls : type (optional, keyword-only)
        Astropy :class:`~astropy.table.Table` class or subclass type to return.
        Default is :class:`~astropy.table.QTable`.
    cosmology_in_meta : bool (optional, keyword-only)
        Whether to put the cosmology class in the Table metadata (if `True`,
        default) or as the first column (if `False`).
    rename : Mapping[str, str] or None (optional, keyword-only)
        A mapping of field names of the |Cosmology| class to column names in the
        |Table|.

    Returns
    -------
    `~astropy.table.QTable`
        With columns for the cosmology parameters, and metadata and
        cosmology class name in the Table's ``meta`` attribute

    Raises
    ------
    TypeError
        If kwarg (optional) 'cls' is not a subclass of `astropy.table.Table`

    Examples
    --------
    A Cosmology as a `~astropy.table.QTable` will have the cosmology's name and
    parameters as columns.

        >>> from astropy.cosmology import Planck18
        >>> ct = Planck18.to_format("astropy.table")
        >>> ct
        <QTable length=1>
          name        H0        Om0    Tcmb0    Neff      m_nu      Ob0
                 km / (Mpc s)            K                 eV
          str8     float64    float64 float64 float64  float64[3] float64
        -------- ------------ ------- ------- ------- ----------- -------
        Planck18        67.66 0.30966  2.7255   3.046 0.0 .. 0.06 0.04897

    The cosmological class and other metadata, e.g. a paper reference, are in
    the Table's metadata.

        >>> ct.meta
        {..., 'cosmology': 'FlatLambdaCDM'}

    To move the cosmology class from the metadata to a Table column, set the
    ``cosmology_in_meta`` argument to `False`:

        >>> Planck18.to_format("astropy.table", cosmology_in_meta=False)
        <QTable length=1>
          cosmology     name        H0        Om0    Tcmb0    Neff      m_nu      Ob0
                               km / (Mpc s)            K                 eV
            str13       str8     float64    float64 float64 float64  float64[3] float64
        ------------- -------- ------------ ------- ------- ------- ----------- -------
        FlatLambdaCDM Planck18        67.66 0.30966  2.7255   3.046 0.0 .. 0.06 0.04897

    Astropy recommends `~astropy.table.QTable` for tables with |Quantity| columns.
    However the returned type may be overridden using the ``cls`` argument:

        >>> from astropy.table import Table
        >>> Planck18.to_format("astropy.table", cls=Table)
        <Table length=1>
        ...

    Fields of the cosmology may be renamed using the ``rename`` argument.

        >>> Planck18.to_format("astropy.table", rename={"H0": "Hubble"})
        <QTable length=1>
          name      Hubble      Om0    Tcmb0    Neff      m_nu      Ob0
                 km / (Mpc s)            K                 eV
          str8     float64    float64 float64 float64  float64[3] float64
        -------- ------------ ------- ------- ------- ----------- -------
        Planck18        67.66 0.30966  2.7255   3.046 0.0 .. 0.06 0.04897
    """
    if not issubclass(cls, Table):
        raise TypeError(f"'cls' must be a (sub)class of Table, not {type(cls)}")

    # Start by getting a map representation.
    data = to_mapping(cosmology)
    data["cosmology"] = data["cosmology"].__qualname__  # change to str

    # Metadata
    meta = data.pop("meta")  # remove the meta
    if cosmology_in_meta:
        meta["cosmology"] = data.pop("cosmology")

    # Need to turn everything into something Table can process:
    # - Column for Parameter
    # - list for anything else
    cosmo_cls = cosmology.__class__
    for k, v in data.items():
        if k in cosmology.parameters:
            col = convert_parameter_to_column(
                cosmo_cls.parameters[k], v, cosmology.meta.get(k)
            )
        else:
            col = Column([v])
        data[k] = col

    tbl = cls(data, meta=meta)

    # Renames
    renames = rename or {}
    for name in tbl.colnames:
        tbl.rename_column(name, renames.get(name, name))

    # Add index
    tbl.add_index(renames.get("name", "name"), unique=True)

    return tbl


def table_identify(
    origin: str, format: str | None, *args: object, **kwargs: object
) -> bool:
    """Identify if object uses the Table format.

    Returns
    -------
    bool
    """
    itis = False
    if origin == "read":
        itis = isinstance(args[1], Table) and (format in (None, "astropy.table"))
    return itis


# ===================================================================
# Register

convert_registry.register_reader("astropy.table", Cosmology, from_table)
convert_registry.register_writer("astropy.table", Cosmology, to_table)
convert_registry.register_identifier("astropy.table", Cosmology, table_identify)



# Licensed under a 3-clause BSD style license - see PYFITS.rst

import io
import math
import os
import pickle
import re
import time
from io import BytesIO

import numpy as np
import pytest
from hypothesis import given
from hypothesis.extra.numpy import basic_indices
from numpy.testing import assert_allclose, assert_equal

from astropy.io import fits
from astropy.io.fits.hdu.compressed import (
    COMPRESSION_TYPES,
    DITHER_SEED_CHECKSUM,
    SUBTRACTIVE_DITHER_1,
)
from astropy.io.fits.tests.conftest import FitsTestCase
from astropy.io.fits.tests.test_table import comparerecords
from astropy.io.fits.verify import VerifyWarning
from astropy.utils.data import download_file
from astropy.utils.misc import NumpyRNGContext


class TestCompressedImage(FitsTestCase):
    def test_empty(self):
        """
        Regression test for https://github.com/astropy/astropy/issues/2595
        """

        hdu = fits.CompImageHDU()
        assert hdu.data is None
        hdu.writeto(self.temp("test.fits"))

        with fits.open(self.temp("test.fits"), mode="update") as hdul:
            assert len(hdul) == 2
            assert isinstance(hdul[1], fits.CompImageHDU)
            assert hdul[1].data is None

            # Now test replacing the empty data with an array and see what
            # happens
            hdul[1].data = np.arange(100, dtype=np.int32)

        with fits.open(self.temp("test.fits")) as hdul:
            assert len(hdul) == 2
            assert isinstance(hdul[1], fits.CompImageHDU)
            assert np.all(hdul[1].data == np.arange(100, dtype=np.int32))

    @pytest.mark.parametrize(
        ("data", "compression_type", "quantize_level"),
        [
            (np.zeros((2, 10, 10), dtype=np.float32), "RICE_1", 16),
            (np.zeros((2, 10, 10), dtype=np.float32), "GZIP_1", -0.01),
            (np.zeros((2, 10, 10), dtype=np.float32), "GZIP_2", -0.01),
            (np.zeros((100, 100)) + 1, "HCOMPRESS_1", 16),
            (np.zeros((10, 10)), "PLIO_1", 16),
        ],
    )
    @pytest.mark.parametrize("byte_order", ["<", ">"])
    def test_comp_image(self, data, compression_type, quantize_level, byte_order):
        data = data.view(data.dtype.newbyteorder(byte_order))
        primary_hdu = fits.PrimaryHDU()
        ofd = fits.HDUList(primary_hdu)
        chdu = fits.CompImageHDU(
            data,
            name="SCI",
            compression_type=compression_type,
            quantize_level=quantize_level,
        )
        ofd.append(chdu)
        ofd.writeto(self.temp("test_new.fits"), overwrite=True)
        ofd.close()
        with fits.open(self.temp("test_new.fits")) as fd:
            assert (fd[1].data == data).all()
            assert fd[1].header["NAXIS"] == chdu.header["NAXIS"]
            assert fd[1].header["NAXIS1"] == chdu.header["NAXIS1"]
            assert fd[1].header["NAXIS2"] == chdu.header["NAXIS2"]
            assert fd[1].header["BITPIX"] == chdu.header["BITPIX"]

    @pytest.mark.remote_data
    def test_comp_image_quantize_level(self):
        """
        Regression test for https://github.com/astropy/astropy/issues/5969

        Test that quantize_level is used.

        """

        np.random.seed(42)

        # Basically what scipy.datasets.ascent() does.
        fname = download_file(
            "https://github.com/scipy/dataset-ascent/blob/main/ascent.dat?raw=true"
        )
        with open(fname, "rb") as f:
            scipy_data = np.array(pickle.load(f))

        data = scipy_data + np.random.randn(512, 512) * 10

        fits.ImageHDU(data).writeto(self.temp("im1.fits"))
        fits.CompImageHDU(
            data,
            compression_type="RICE_1",
            quantize_method=1,
            quantize_level=-1,
            dither_seed=5,
        ).writeto(self.temp("im2.fits"))
        fits.CompImageHDU(
            data,
            compression_type="RICE_1",
            quantize_method=1,
            quantize_level=-100,
            dither_seed=5,
        ).writeto(self.temp("im3.fits"))

        im1 = fits.getdata(self.temp("im1.fits"))
        im2 = fits.getdata(self.temp("im2.fits"))
        im3 = fits.getdata(self.temp("im3.fits"))

        assert not np.array_equal(im2, im3)
        assert np.isclose(np.min(im1 - im2), -0.5, atol=1e-3)
        assert np.isclose(np.max(im1 - im2), 0.5, atol=1e-3)
        assert np.isclose(np.min(im1 - im3), -50, atol=1e-1)
        assert np.isclose(np.max(im1 - im3), 50, atol=1e-1)

    def test_comp_image_hcompression_1_invalid_data(self):
        """
        Tests compression with the HCOMPRESS_1 algorithm with data that is
        not 2D and has a non-2D tile size.
        """

        pytest.raises(
            ValueError,
            fits.CompImageHDU,
            np.zeros((2, 10, 10), dtype=np.float32),
            name="SCI",
            compression_type="HCOMPRESS_1",
            quantize_level=16,
            tile_shape=(2, 10, 10),
        )

    def test_comp_image_hcompress_image_stack(self):
        """
        Regression test for https://aeon.stsci.edu/ssb/trac/pyfits/ticket/171

        Tests that data containing more than two dimensions can be
        compressed with HCOMPRESS_1 so long as the user-supplied tile size can
        be flattened to two dimensions.
        """

        cube = np.arange(300, dtype=np.float32).reshape(3, 10, 10)
        hdu = fits.CompImageHDU(
            data=cube,
            name="SCI",
            compression_type="HCOMPRESS_1",
            quantize_level=16,
            tile_shape=(1, 5, 5),
        )
        hdu.writeto(self.temp("test.fits"))

        with fits.open(self.temp("test.fits")) as hdul:
            # HCOMPRESSed images are allowed to deviate from the original by
            # about 1/quantize_level of the RMS in each tile.
            assert np.abs(hdul["SCI"].data - cube).max() < 1.0 / 15.0

    def test_subtractive_dither_seed(self):
        """
        Regression test for https://github.com/spacetelescope/PyFITS/issues/32

        Ensure that when floating point data is compressed with the
        SUBTRACTIVE_DITHER_1 quantization method that the correct ZDITHER0 seed
        is added to the header, and that the data can be correctly
        decompressed.
        """

        array = np.arange(100.0).reshape(10, 10)
        csum = (array[0].view("uint8").sum() % 10000) + 1
        hdu = fits.CompImageHDU(
            data=array,
            quantize_method=SUBTRACTIVE_DITHER_1,
            dither_seed=DITHER_SEED_CHECKSUM,
        )
        hdu.writeto(self.temp("test.fits"))

        with fits.open(self.temp("test.fits")) as hdul:
            comp_header = hdul[1]._bintable.header
            assert "ZQUANTIZ" in comp_header
            assert comp_header["ZQUANTIZ"] == "SUBTRACTIVE_DITHER_1"
            assert "ZDITHER0" in comp_header
            assert comp_header["ZDITHER0"] == csum
            assert np.all(hdul[1].data == array)

    def test_disable_image_compression(self):
        with fits.open(self.data("comp.fits"), disable_image_compression=True) as hdul:
            # The compressed image HDU should show up as a BinTableHDU, but
            # *not* a CompImageHDU
            assert isinstance(hdul[1], fits.BinTableHDU)
            assert not isinstance(hdul[1], fits.CompImageHDU)

        with fits.open(self.data("comp.fits")) as hdul:
            assert isinstance(hdul[1], fits.CompImageHDU)

    def test_open_comp_image_in_update_mode(self):
        """
        Regression test for https://aeon.stsci.edu/ssb/trac/pyfits/ticket/167

        Similar to test_open_scaled_in_update_mode(), but specifically for
        compressed images.
        """

        # Copy the original file before making any possible changes to it
        testfile = self.copy_file("comp.fits")
        mtime = os.stat(testfile).st_mtime

        time.sleep(1)

        fits.open(testfile, mode="update").close()

        # Ensure that no changes were made to the file merely by immediately
        # opening and closing it.
        assert mtime == os.stat(testfile).st_mtime

    @pytest.mark.slow
    def test_open_scaled_in_update_mode_compressed(self):
        """
        Regression test for https://aeon.stsci.edu/ssb/trac/pyfits/ticket/88 2

        Identical to test_open_scaled_in_update_mode() but with a compressed
        version of the scaled image.
        """

        # Copy+compress the original file before making any possible changes to
        # it
        with fits.open(self.data("scale.fits"), do_not_scale_image_data=True) as hdul:
            chdu = fits.CompImageHDU(data=hdul[0].data, header=hdul[0].header)
            chdu.header["BZERO"] = hdul[0].header["BZERO"]
            chdu.header["BSCALE"] = hdul[0].header["BSCALE"]
            chdu.writeto(self.temp("scale.fits"))
        mtime = os.stat(self.temp("scale.fits")).st_mtime

        time.sleep(1)

        # Now open the file in update mode and close immediately. Note that we
        # need to set do_not_scale_image_data otherwise the data is scaled upon
        # being opened.
        fits.open(
            self.temp("scale.fits"),
            mode="update",
            do_not_scale_image_data=True,
        ).close()

        # Ensure that no changes were made to the file merely by immediately
        # opening and closing it.
        assert mtime == os.stat(self.temp("scale.fits")).st_mtime

        # Insert a slight delay to ensure the mtime does change when the file
        # is changed
        time.sleep(1)

        hdul = fits.open(self.temp("scale.fits"), "update")
        hdul[1].data
        hdul.close()

        # Now the file should be updated with the rescaled data
        assert mtime != os.stat(self.temp("scale.fits")).st_mtime
        hdul = fits.open(self.temp("scale.fits"), mode="update")
        assert hdul[1].data.dtype == np.dtype("float32")
        assert hdul[1].header["BITPIX"] == -32
        assert "BZERO" not in hdul[1].header
        assert "BSCALE" not in hdul[1].header

        # Try reshaping the data, then closing and reopening the file; let's
        # see if all the changes are preserved properly
        hdul[1].data.shape = (42, 10)
        hdul.close()

        hdul = fits.open(self.temp("scale.fits"))
        assert hdul[1].shape == (42, 10)
        assert hdul[1].data.dtype == np.dtype("float32")
        assert hdul[1].header["BITPIX"] == -32
        assert "BZERO" not in hdul[1].header
        assert "BSCALE" not in hdul[1].header
        hdul.close()

    def test_write_comp_hdu_direct_from_existing(self):
        with fits.open(self.data("comp.fits")) as hdul:
            hdul[1].writeto(self.temp("test.fits"))

        with fits.open(self.data("comp.fits")) as hdul1:
            with fits.open(self.temp("test.fits")) as hdul2:
                assert np.all(hdul1[1].data == hdul2[1].data)
                assert comparerecords(
                    hdul1[1].compressed_data, hdul2[1].compressed_data
                )

    def test_rewriting_large_scaled_image_compressed(self):
        """
        Regression test for https://aeon.stsci.edu/ssb/trac/pyfits/ticket/88 1

        Identical to test_rewriting_large_scaled_image() but with a compressed
        image.
        """

        with fits.open(
            self.data("fixed-1890.fits"), do_not_scale_image_data=True
        ) as hdul:
            chdu = fits.CompImageHDU(data=hdul[0].data, header=hdul[0].header)
            chdu.writeto(self.temp("fixed-1890-z.fits"))

        hdul = fits.open(self.temp("fixed-1890-z.fits"))
        orig_data = hdul[1].data
        hdul.writeto(self.temp("test_new.fits"), overwrite=True)
        hdul.close()
        hdul = fits.open(self.temp("test_new.fits"))
        assert (hdul[1].data == orig_data).all()
        hdul.close()

        # Just as before, but this time don't touch hdul[0].data before writing
        # back out--this is the case that failed in
        # https://aeon.stsci.edu/ssb/trac/pyfits/ticket/84
        hdul = fits.open(self.temp("fixed-1890-z.fits"))
        hdul.writeto(self.temp("test_new.fits"), overwrite=True)
        hdul.close()
        hdul = fits.open(self.temp("test_new.fits"))
        assert (hdul[1].data == orig_data).all()
        hdul.close()

        # Test opening/closing/reopening a scaled file in update mode
        hdul = fits.open(self.temp("fixed-1890-z.fits"), do_not_scale_image_data=True)
        hdul.writeto(
            self.temp("test_new.fits"), overwrite=True, output_verify="silentfix"
        )
        hdul.close()
        hdul = fits.open(self.temp("test_new.fits"))
        orig_data = hdul[1].data
        hdul.close()
        hdul = fits.open(self.temp("test_new.fits"), mode="update")
        hdul.close()
        hdul = fits.open(self.temp("test_new.fits"))
        assert (hdul[1].data == orig_data).all()
        hdul.close()

    def test_scale_back_compressed(self):
        """
        Regression test for https://aeon.stsci.edu/ssb/trac/pyfits/ticket/88 3

        Identical to test_scale_back() but uses a compressed image.
        """

        # Create a compressed version of the scaled image
        with fits.open(self.data("scale.fits"), do_not_scale_image_data=True) as hdul:
            chdu = fits.CompImageHDU(data=hdul[0].data, header=hdul[0].header)
            chdu.header["BZERO"] = hdul[0].header["BZERO"]
            chdu.header["BSCALE"] = hdul[0].header["BSCALE"]
            chdu.writeto(self.temp("scale.fits"))

        with fits.open(self.temp("scale.fits"), mode="update", scale_back=True) as hdul:
            orig_bitpix = hdul[1].header["BITPIX"]
            orig_bzero = hdul[1].header["BZERO"]
            orig_bscale = hdul[1].header["BSCALE"]
            orig_data = hdul[1].data.copy()
            hdul[1].data[0] = 0

        with fits.open(self.temp("scale.fits"), do_not_scale_image_data=True) as hdul:
            assert hdul[1].header["BITPIX"] == orig_bitpix
            assert hdul[1].header["BZERO"] == orig_bzero
            assert hdul[1].header["BSCALE"] == orig_bscale

            zero_point = math.floor(-orig_bzero / orig_bscale)
            assert (hdul[1].data[0] == zero_point).all()

        with fits.open(self.temp("scale.fits")) as hdul:
            assert_equal(hdul[1].data[1:], orig_data[1:])
            # Extra test to ensure that after everything the data is still the
            # same as in the original uncompressed version of the image
            with fits.open(self.data("scale.fits")) as hdul2:
                # Recall we made the same modification to the data in hdul
                # above
                hdul2[0].data[0] = 0
                assert_equal(hdul[1].data, hdul2[0].data)

    def test_lossless_gzip_compression(self):
        """Regression test for https://aeon.stsci.edu/ssb/trac/pyfits/ticket/198"""

        rng = np.random.default_rng(42)
        noise = rng.normal(size=(20, 20))

        chdu1 = fits.CompImageHDU(data=noise, compression_type="GZIP_1")
        # First make a test image with lossy compression and make sure it
        # wasn't compressed perfectly.  This shouldn't happen ever, but just to
        # make sure the test non-trivial.
        chdu1.writeto(self.temp("test.fits"))

        with fits.open(self.temp("test.fits")) as h:
            assert np.abs(noise - h[1].data).max() > 0.0

        del h

        chdu2 = fits.CompImageHDU(
            data=noise, compression_type="GZIP_1", quantize_level=0.0
        )  # No quantization
        chdu2.writeto(self.temp("test.fits"), overwrite=True)

        with fits.open(self.temp("test.fits")) as h:
            assert (noise == h[1].data).all()

    def test_compression_column_tforms(self):
        """Regression test for https://aeon.stsci.edu/ssb/trac/pyfits/ticket/199"""

        # Some interestingly tiled data so that some of it is quantized and
        # some of it ends up just getting gzip-compressed
        data2 = (np.arange(1, 8, dtype=np.float32) * 10)[:, np.newaxis] + np.arange(
            1, 7
        )
        np.random.seed(1337)
        data1 = np.random.uniform(size=(6 * 4, 7 * 4))
        data1[: data2.shape[0], : data2.shape[1]] = data2
        chdu = fits.CompImageHDU(data1, compression_type="RICE_1", tile_shape=(6, 7))
        chdu.writeto(self.temp("test.fits"))

        with fits.open(self.temp("test.fits"), disable_image_compression=True) as h:
            assert re.match(r"^1PB\(\d+\)$", h[1].header["TFORM1"])
            assert re.match(r"^1PB\(\d+\)$", h[1].header["TFORM2"])

    def test_compression_update_header(self):
        """Regression test for
        https://github.com/spacetelescope/PyFITS/issues/23
        """

        testfile = self.copy_file("comp.fits")
        with fits.open(testfile, mode="update") as hdul:
            assert isinstance(hdul[1], fits.CompImageHDU)
            hdul[1].header["test1"] = "test"
            hdul[1]._header["test2"] = "test2"

        with fits.open(testfile) as hdul:
            assert "test1" in hdul[1].header
            assert hdul[1].header["test1"] == "test"
            assert "test2" in hdul[1].header
            assert hdul[1].header["test2"] == "test2"

        # Test update via index now:
        with fits.open(testfile, mode="update") as hdul:
            hdr = hdul[1].header
            hdr[hdr.index("TEST1")] = "foo"

        with fits.open(testfile) as hdul:
            assert hdul[1].header["TEST1"] == "foo"

        # Test slice updates
        with fits.open(testfile, mode="update") as hdul:
            hdul[1].header["TEST*"] = "qux"

        with fits.open(testfile) as hdul:
            assert list(hdul[1].header["TEST*"].values()) == ["qux", "qux"]

        with fits.open(testfile, mode="update") as hdul:
            hdr = hdul[1].header
            idx = hdr.index("TEST1")
            hdr[idx : idx + 2] = "bar"

        with fits.open(testfile) as hdul:
            assert list(hdul[1].header["TEST*"].values()) == ["bar", "bar"]

        # Test updating a specific COMMENT card duplicate
        with fits.open(testfile, mode="update") as hdul:
            hdul[1].header[("COMMENT", 1)] = "I am fire. I am death!"

        with fits.open(testfile) as hdul:
            assert hdul[1].header["COMMENT"][1] == "I am fire. I am death!"
            assert hdul[1]._header["COMMENT"][1] == "I am fire. I am death!"

        # Test deleting by keyword and by slice
        with fits.open(testfile, mode="update") as hdul:
            hdr = hdul[1].header
            del hdr["COMMENT"]
            idx = hdr.index("TEST1")
            del hdr[idx : idx + 2]

        with fits.open(testfile) as hdul:
            assert "COMMENT" not in hdul[1].header
            assert "COMMENT" not in hdul[1]._header
            assert "TEST1" not in hdul[1].header
            assert "TEST1" not in hdul[1]._header
            assert "TEST2" not in hdul[1].header
            assert "TEST2" not in hdul[1]._header

    def test_compression_update_header_with_reserved(self, tmp_path):
        """
        Ensure that setting reserved keywords related to the table data
        structure on CompImageHDU image headers fails.
        """
        with fits.open(self.data("comp.fits")) as hdul:
            hdr = hdul[1].header
            hdr["TFIELDS"] = 8
            hdr["THEAP"] = 1000
            hdr["TTYPE1"] = "Foo"
            hdr["ZCMPTYPE"] = "ASDF"
            hdr["ZVAL1"] = "Foo"
            with pytest.warns() as record:
                hdul.writeto(tmp_path / "test.fits")
            assert len(record) == 5
            for i, keyword in enumerate(
                ("TFIELDS", "THEAP", "TTYPE1", "ZCMPTYPE", "ZVAL1")
            ):
                assert f"Keyword {keyword!r} is reserved" in record[i].message.args[0]

    def test_compression_header_append(self, tmp_path):
        with fits.open(self.data("comp.fits")) as hdul:
            imghdr = hdul[1].header

            imghdr.append("TFIELDS")

            imghdr.append(("FOO", "bar", "qux"), end=True)
            assert "FOO" in imghdr
            assert imghdr[-1] == "bar"

            imghdr.append(("CHECKSUM", "abcd1234"))
            assert "CHECKSUM" in imghdr
            assert imghdr["CHECKSUM"] == "abcd1234"

            with pytest.warns(
                VerifyWarning, match="Keyword 'TFIELDS' is reserved"
            ) as w:
                hdul.writeto(tmp_path / "updated.fits")

        with fits.open(
            tmp_path / "updated.fits", disable_image_compression=True
        ) as hdulc:
            tblhdr = hdulc[1].header

            assert "FOO" in tblhdr
            assert tblhdr["FOO"] == "bar"

            assert "CHECKSUM" not in tblhdr
            assert "ZHECKSUM" in tblhdr
            assert tblhdr["ZHECKSUM"] == "abcd1234"

    def test_compression_header_append2(self):
        """
        Regression test for issue https://github.com/astropy/astropy/issues/5827
        """
        with fits.open(self.data("comp.fits")) as hdul:
            header = hdul[1].header
            while len(header) < 1000:
                header.append()  # pad with grow room

            # Append stats to header:
            header.append(("Q1_OSAVG", 1, "[adu] quadrant 1 overscan mean"))
            header.append(("Q1_OSSTD", 1, "[adu] quadrant 1 overscan stddev"))
            header.append(("Q1_OSMED", 1, "[adu] quadrant 1 overscan median"))

    def test_compression_header_insert(self, tmp_path):
        with fits.open(self.data("comp.fits")) as hdul:
            imghdr = hdul[1].header

            # First try inserting a restricted keyword
            imghdr.insert(1000, "TFIELDS")

            # First try keyword-relative insert
            imghdr.insert("TELESCOP", ("OBSERVER", "Phil Plait"))
            assert "OBSERVER" in imghdr
            assert imghdr.index("OBSERVER") == imghdr.index("TELESCOP") - 1

            # Next let's see if an index-relative insert winds up being
            # sensible
            idx = imghdr.index("OBSERVER")
            imghdr.insert("OBSERVER", ("FOO",))
            assert "FOO" in imghdr
            assert imghdr.index("FOO") == idx

            with pytest.warns(
                VerifyWarning, match="Keyword 'TFIELDS' is reserved"
            ) as w:
                hdul.writeto(tmp_path / "updated.fits")

        with fits.open(
            tmp_path / "updated.fits", disable_image_compression=True
        ) as hdulc:
            tblhdr = hdulc[1].header

            assert tblhdr.count("TFIELDS") == 1

            assert "OBSERVER" in tblhdr
            assert tblhdr.index("OBSERVER") == tblhdr.index("TELESCOP") - 1

            assert "FOO" in tblhdr
            assert tblhdr.index("FOO") == tblhdr.index("OBSERVER") - 1

    def test_compression_header_set_before_after(self, tmp_path):
        with fits.open(self.data("comp.fits")) as hdul:
            imghdr = hdul[1].header

            imghdr.set("ZBITPIX", 77, "asdf", after="GCOUNT")

            with pytest.warns(UserWarning, match="Keyword 'ZBITPIX' is reserved"):
                hdul.writeto(tmp_path / "updated1.fits")

        with fits.open(
            tmp_path / "updated1.fits", disable_image_compression=True
        ) as hdulc:
            tblhdr = hdulc[1].header

            assert tblhdr.count("ZBITPIX") == 1
            assert tblhdr["ZBITPIX"] != 77

        with fits.open(self.data("comp.fits")) as hdul:
            imghdr = hdul[1].header

            imghdr.set("FOO", 2, before="OBJECT")
            imghdr.set("BAR", 3, after="OBJECT")
            assert imghdr.index("FOO") == imghdr.index("OBJECT") - 1
            assert imghdr.index("BAR") == imghdr.index("OBJECT") + 1
            assert imghdr["FOO"] == 2
            assert imghdr["BAR"] == 3

            hdul.writeto(tmp_path / "updated2.fits")

        with fits.open(
            tmp_path / "updated2.fits", disable_image_compression=True
        ) as hdulc:
            tblhdr = hdulc[1].header

            assert tblhdr.index("FOO") == tblhdr.index("OBJECT") - 1
            assert tblhdr.index("BAR") == tblhdr.index("OBJECT") + 1
            assert tblhdr["FOO"] == 2
            assert tblhdr["BAR"] == 3

    def test_compression_header_append_commentary(self):
        """
        Regression test for https://github.com/astropy/astropy/issues/2363
        """

        hdu = fits.CompImageHDU(np.array([0], dtype=np.int32))
        hdu.header["COMMENT"] = "hello world"
        assert hdu.header["COMMENT"] == ["hello world"]
        hdu.writeto(self.temp("test.fits"))

        with fits.open(self.temp("test.fits")) as hdul:
            assert hdul[1].header["COMMENT"] == ["hello world"]

    def test_compression_with_gzip_column(self):
        """
        Regression test for https://github.com/spacetelescope/PyFITS/issues/71
        """

        arr = np.zeros((2, 7000), dtype="float32")

        # The first row (which will be the first compressed tile) has a very
        # wide range of values that will be difficult to quantize, and should
        # result in use of a GZIP_COMPRESSED_DATA column
        arr[0] = np.linspace(0, 1, 7000)
        arr[1] = np.random.normal(size=7000)

        hdu = fits.CompImageHDU(data=arr)
        hdu.writeto(self.temp("test.fits"))

        with fits.open(self.temp("test.fits")) as hdul:
            comp_hdu = hdul[1]

            # GZIP-compressed tile should compare exactly
            assert np.all(comp_hdu.data[0] == arr[0])
            # The second tile uses lossy compression and may be somewhat off,
            # so we don't bother comparing it exactly

    def test_duplicate_compression_header_keywords(self):
        """
        Regression test for https://github.com/astropy/astropy/issues/2750

        Tests that the fake header (for the compressed image) can still be read
        even if the real header contained a duplicate ZTENSION keyword (the
        issue applies to any keyword specific to the compression convention,
        however).
        """

        arr = np.arange(100, dtype=np.int32)
        hdu = fits.CompImageHDU(data=arr)
        hdu.writeto(self.temp("test1.fits"))

        # append the duplicate keyword
        with fits.open(
            self.temp("test1.fits"), disable_image_compression=True
        ) as hdulc:
            hdulc[1].header.append(("ZTENSION", "IMAGE"))
            header = hdulc[1].header
            hdulc.writeto(self.temp("test2.fits"))

        with fits.open(self.temp("test2.fits")) as hdul:
            assert header == hdul[1]._bintable.header
            # There's no good reason to have a duplicate keyword, but
            # technically it isn't invalid either :/
            assert hdul[1]._bintable.header.count("ZTENSION") == 2

    def test_scale_bzero_with_compressed_int_data(self):
        """
        Regression test for https://github.com/astropy/astropy/issues/4600
        and https://github.com/astropy/astropy/issues/4588

        Identical to test_scale_bzero_with_int_data() but uses a compressed
        image.
        """

        a = np.arange(100, 200, dtype=np.int16)

        hdu1 = fits.CompImageHDU(data=a.copy())
        hdu2 = fits.CompImageHDU(data=a.copy())
        # Previously the following line would throw a TypeError,
        # now it should be identical to the integer bzero case
        hdu1.scale("int16", bzero=99.0)
        hdu2.scale("int16", bzero=99)
        assert np.allclose(hdu1.data, hdu2.data)

    def test_scale_back_compressed_uint_assignment(self):
        """
        Extend fix for #4600 to assignment to data

        Identical to test_scale_back_uint_assignment() but uses a compressed
        image.

        Suggested by:
        https://github.com/astropy/astropy/pull/4602#issuecomment-208713748
        """

        a = np.arange(100, 200, dtype=np.uint16)
        fits.CompImageHDU(a).writeto(self.temp("test.fits"))
        with fits.open(self.temp("test.fits"), mode="update", scale_back=True) as hdul:
            hdul[1].data[:] = 0
            assert np.allclose(hdul[1].data, 0)

    def test_compressed_header_double_extname(self):
        """Test that a double EXTNAME with one default value does not
        mask the non-default value."""
        with fits.open(self.data("double_ext.fits")) as hdul:
            hdu = hdul[1]

            # The non-default name should be returned.
            assert hdu.name == "ccd00"
            assert "EXTNAME" in hdu.header
            assert hdu.name == hdu.header["EXTNAME"]

            # There should be 1 non-default EXTNAME entries.
            indices = hdu.header._keyword_indices["EXTNAME"]
            assert len(indices) == 1

            # Test header sync from property set.
            new_name = "NEW_NAME"
            hdu.name = new_name
            assert hdu.name == new_name
            assert hdu.header["EXTNAME"] == new_name

            # Check that setting the header will change the name property.
            hdu.header["EXTNAME"] = "NEW2"
            assert hdu.name == "NEW2"

            hdul.writeto(self.temp("tmp.fits"), overwrite=True)
            with fits.open(self.temp("tmp.fits")) as hdul1:
                hdu1 = hdul1[1]
                assert len(hdu.header._keyword_indices["EXTNAME"]) == 1
                assert hdu1.name == "NEW2"

            # Check that deleting EXTNAME will and setting the name will
            # work properly.
            del hdu.header["EXTNAME"]
            hdu.name = "RE-ADDED"
            assert hdu.name == "RE-ADDED"

            with pytest.raises(TypeError):
                hdu.name = 42

    def test_compressed_header_extname(self):
        """Test consistent EXTNAME / hdu name interaction."""
        name = "FOO"
        hdu = fits.CompImageHDU(data=np.arange(10), name=name)
        assert hdu._header["EXTNAME"] == name
        assert hdu.header["EXTNAME"] == name
        assert hdu.name == name

        name = "BAR"
        hdu.name = name
        assert hdu._header["EXTNAME"] == name
        assert hdu.header["EXTNAME"] == name
        assert hdu.name == name

        assert len(hdu._header._keyword_indices["EXTNAME"]) == 1

    def test_compressed_header_minimal(self):
        """
        Regression test for https://github.com/astropy/astropy/issues/11694

        Tests that CompImageHDU can be initialized with a Header that
        contains few or no cards, and doesn't require specific cards
        such as 'BITPIX' or 'NAXIS'.
        """
        fits.CompImageHDU(data=np.arange(10), header=fits.Header())
        header = fits.Header({"HELLO": "world"})
        hdu = fits.CompImageHDU(data=np.arange(10), header=header)
        assert hdu.header["HELLO"] == "world"

    @pytest.mark.parametrize(
        ("keyword", "dtype", "expected"),
        [
            ("BSCALE", np.uint8, np.float32),
            ("BSCALE", np.int16, np.float32),
            ("BSCALE", np.int32, np.float64),
            ("BZERO", np.uint8, np.float32),
            ("BZERO", np.int16, np.float32),
            ("BZERO", np.int32, np.float64),
        ],
    )
    def test_compressed_scaled_float(self, keyword, dtype, expected):
        """
        If BSCALE,BZERO is set to floating point values, the image
        should be floating-point.

        https://github.com/astropy/astropy/pull/6492

        Parameters
        ----------
        keyword : `str`
            Keyword to set to a floating-point value to trigger
            floating-point pixels.
        dtype : `numpy.dtype`
            Type of original array.
        expected : `numpy.dtype`
            Expected type of uncompressed array.
        """
        value = 1.23345  # A floating-point value
        hdu = fits.CompImageHDU(np.arange(0, 10, dtype=dtype))
        hdu.header[keyword] = value
        hdu.writeto(self.temp("test.fits"))
        del hdu
        with fits.open(self.temp("test.fits")) as hdu:
            assert hdu[1].header[keyword] == value
            assert hdu[1].data.dtype == expected

    @pytest.mark.parametrize(
        "dtype", (np.uint8, np.int16, np.uint16, np.int32, np.uint32)
    )
    def test_compressed_integers(self, dtype):
        """Test that the various integer dtypes are correctly written and read.

        Regression test for https://github.com/astropy/astropy/issues/9072

        """
        mid = np.iinfo(dtype).max // 2
        data = np.arange(mid - 50, mid + 50, dtype=dtype)
        testfile = self.temp("test.fits")
        hdu = fits.CompImageHDU(data=data)
        hdu.writeto(testfile, overwrite=True)
        new = fits.getdata(testfile)
        np.testing.assert_array_equal(data, new)

    @pytest.mark.parametrize("dtype", ["f", "i4"])
    @pytest.mark.parametrize("compression_type", COMPRESSION_TYPES)
    def test_write_non_contiguous_data(self, dtype, compression_type):
        """
        Regression test for https://github.com/astropy/astropy/issues/2150

        This used to require changing the whole array to be C-contiguous before
        passing to CFITSIO, but we no longer need this - our explicit conversion
        to bytes in the compression codecs returns contiguous bytes for each
        tile on-the-fly.
        """

        orig = np.arange(400, dtype=dtype).reshape((20, 20), order="f")[::2, ::2]
        assert not orig.flags.contiguous
        primary = fits.PrimaryHDU()
        hdu = fits.CompImageHDU(orig, compression_type=compression_type)
        hdulist = fits.HDUList([primary, hdu])
        hdulist.writeto(self.temp("test.fits"))

        actual = fits.getdata(self.temp("test.fits"))
        assert_equal(orig, actual)

    def test_slice_and_write_comp_hdu(self):
        """
        Regression test for https://github.com/astropy/astropy/issues/9955
        """
        with fits.open(self.data("comp.fits")) as hdul:
            hdul[1].data = hdul[1].data[:200, :100]
            assert not hdul[1].data.flags.contiguous
            hdul[1].writeto(self.temp("test.fits"))

        with fits.open(self.data("comp.fits")) as hdul1:
            with fits.open(self.temp("test.fits")) as hdul2:
                assert_equal(hdul1[1].data[:200, :100], hdul2[1].data)

    def test_comp_image_properties_default(self):
        chdu = fits.CompImageHDU(np.zeros((3, 4, 5)))
        assert chdu.tile_shape == (1, 1, 5)
        assert chdu.compression_type == "RICE_1"

    def test_comp_image_properties_set(self):
        chdu = fits.CompImageHDU(
            np.zeros((3, 4, 5)), compression_type="PLIO_1", tile_shape=(2, 3, 4)
        )
        assert chdu.tile_shape == (2, 3, 4)
        assert chdu.compression_type == "PLIO_1"

    def test_compressed_optional_prefix_tform(self, tmp_path):
        # Regression test for a bug that caused an error if a
        # compressed file had TFORM missing the optional 1 prefix

        data = np.zeros((3, 4, 5))

        hdu1 = fits.CompImageHDU(data=data)
        hdu1.writeto(tmp_path / "compressed.fits")

        with fits.open(
            tmp_path / "compressed.fits", disable_image_compression=True, mode="update"
        ) as hdul:
            assert hdul[1].header["TFORM1"] == "1PB(0)"
            assert hdul[1].header["TFORM2"] == "1PB(24)"
            hdul[1].header["TFORM1"] = "PB(0)"
            hdul[1].header["TFORM2"] = "PB(24)"

        with fits.open(
            tmp_path / "compressed.fits", disable_image_compression=True
        ) as hdul:
            assert hdul[1].header["TFORM1"] == "PB(0)"
            assert hdul[1].header["TFORM2"] == "PB(24)"

        with fits.open(tmp_path / "compressed.fits") as hdul:
            assert_equal(hdul[1].data, data)

    def test_info(self):
        """
        Make sure .info() works correctly when CompImageHDUs are present
        """
        output = io.StringIO()
        with fits.open(self.data("comp.fits")) as hdul:
            hdul.info(output=output)
        output.seek(0)

        # Note: ignore the first line which just gives the filename
        actual = output.read().splitlines()[1:]

        expected = [
            "No.    Name      Ver    Type      Cards   Dimensions   Format",
            "0  PRIMARY       1 PrimaryHDU       4   ()",
            "1  COMPRESSED_IMAGE    1 CompImageHDU    105   (440, 300)   int16",
        ]

        for line_actual, line_expected in zip(actual, expected, strict=True):
            assert line_actual.strip() == line_expected.strip()

    def test_shape(self):
        with fits.open(self.data("comp.fits")) as hdul:
            assert hdul[1].header["NAXIS1"] == 440
            assert hdul[1].header["NAXIS2"] == 300
            assert hdul[1].shape == (300, 440)
            hdul[1].data = np.ones((120, 150))
            assert hdul[1].header["NAXIS1"] == 150
            assert hdul[1].header["NAXIS2"] == 120
            assert hdul[1].shape == (120, 150)

    def test_inplace_data_modify(self, tmp_path):
        testfile = self.copy_file("comp.fits")

        with fits.open(testfile, mode="update") as hdul:
            data = hdul[1].data
            data[0] = 0

        with fits.open(testfile) as hdul:
            assert hdul[1].data[0, 0] == 0

    def test_summary_noload(self):
        # Make sure that calling info() (and therefore CompImageHDU.summary)
        # does not cause the data to be loaded.
        with fits.open(self.data("comp.fits")) as hdul:
            summary = hdul.info(output=False)
            assert summary == [
                (0, "PRIMARY", 1, "PrimaryHDU", 4, (), "", ""),
                (
                    1,
                    "COMPRESSED_IMAGE",
                    1,
                    "CompImageHDU",
                    105,
                    (440, 300),
                    "int16",
                    "",
                ),
            ]
            assert not hdul[1]._data_loaded

    def test_fileinfo(self):
        with fits.open(self.data("comp.fits")) as hdul:
            res = hdul.fileinfo(1)

        assert res["datLoc"] == 14400
        assert res["datSpan"] == 72000
        assert res["filemode"] == "readonly"
        assert res["filename"] == self.data("comp.fits")
        assert res["hdrLoc"] == 2880
        assert not res["resized"]


class TestCompHDUSections:
    @pytest.fixture(autouse=True)
    def setup_method(self, tmp_path):
        shape = (13, 17, 25)
        self.data = np.arange(np.prod(shape)).reshape(shape).astype(np.int32)

        header1 = fits.Header()
        hdu1 = fits.CompImageHDU(
            self.data, header1, compression_type="RICE_1", tile_shape=(5, 4, 5)
        )

        header2 = fits.Header()
        hdu2 = fits.CompImageHDU(
            self.data, header2, compression_type="RICE_1", tile_shape=(5, 4, 5)
        )
        hdu2.header["BSCALE"] = 2
        hdu2.header["BZERO"] = 100
        hdulist = fits.HDUList([fits.PrimaryHDU(), hdu1, hdu2])
        hdulist.writeto(tmp_path / "sections.fits")

        self.hdul = fits.open(tmp_path / "sections.fits")
        self.hdul2 = fits.open(tmp_path / "sections.fits")

    def teardown_method(self):
        self.hdul.close()
        self.hdul = None

    @given(basic_indices((13, 17, 25)))
    def test_section_slicing(self, index):
        assert_equal(self.hdul[1].section[index], self.hdul[1].data[index])
        assert_equal(self.hdul[1].section[index], self.data[index])

    @given(basic_indices((13, 17, 25)))
    def test_section_slicing_scaling(self, index):
        assert_equal(self.hdul[2].section[index], self.hdul[2].data[index])
        assert_equal(self.hdul[2].section[index], self.data[index] * 2 + 100)

    def test_section_properties(self):
        assert self.hdul[1].section.dtype is np.dtype("int32")
        assert self.hdul[1].section.ndim == 3


def test_comphdu_fileobj():
    # Regression test for a bug that caused an error to happen
    # internally when reading the data if requested data shapes
    # were not plain integers - this was triggered when accessing
    # sections on data backed by certain kinds of objects such as
    # BytesIO (but not regular file handles)

    data = np.arange(6).reshape((2, 3)).astype(np.int32)

    byte_buffer = BytesIO()

    header = fits.Header()
    hdu = fits.CompImageHDU(data, header, compression_type="RICE_1")
    hdu.writeto(byte_buffer)

    byte_buffer.seek(0)

    hdu2 = fits.open(byte_buffer, mode="readonly")[1]
    assert hdu2.section[1, 2] == 5


def test_comphdu_bscale(tmp_path):
    """
    Regression test for a bug that caused extensions that used BZERO and BSCALE
    that got turned into CompImageHDU to end up with BZERO/BSCALE before the
    TFIELDS.
    """

    filename1 = tmp_path / "3hdus.fits"
    filename2 = tmp_path / "3hdus_comp.fits"

    x = np.random.random((100, 100)) * 100

    x0 = fits.PrimaryHDU()
    x1 = fits.ImageHDU(np.array(x - 50, dtype=int), uint=True)
    x1.header["BZERO"] = 20331
    x1.header["BSCALE"] = 2.3
    hdus = fits.HDUList([x0, x1])
    hdus.writeto(filename1)

    # fitsverify (based on cfitsio) should fail on this file, only seeing the
    # first HDU.
    with fits.open(filename1) as hdus:
        hdus[1] = fits.CompImageHDU(
            data=hdus[1].data.astype(np.uint32), header=hdus[1].header
        )
        hdus.writeto(filename2)

    # open again and verify
    with fits.open(filename2) as hdus:
        hdus[1].verify("exception")


def test_image_write_readonly(tmp_path):
    # Regression test to make sure that we can write out read-only arrays (#5512)

    x = np.array([1.0, 2.0, 3.0])
    x.setflags(write=False)
    ghdu = fits.CompImageHDU(data=x)
    # add_datasum does not work for CompImageHDU
    # ghdu.add_datasum()

    filename = tmp_path / "test2.fits"

    ghdu.writeto(filename)

    with fits.open(filename) as hdulist:
        assert_equal(hdulist[1].data, [1.0, 2.0, 3.0])


def test_uint_option(tmp_path):
    """
    Check that the uint= option works correctly
    """

    filename = tmp_path / "uint_test.fits"

    data = (2 ** (1 + np.arange(16).reshape((4, 4))) - 1).astype(np.uint16)

    hdu = fits.CompImageHDU(data)
    hdu.writeto(filename)

    with fits.open(filename) as hdulist:
        assert hdulist[1].data.dtype == np.dtype("uint16")
        assert_equal(hdulist[1].data, data)

    with fits.open(filename, uint=False) as hdulist:
        assert hdulist[1].data.dtype == np.dtype("float32")
        assert_allclose(hdulist[1].data, data)


def test_incorrect_bzero(tmp_path):
    """
    Regression test for https://github.com/astropy/astropy/issues/5999 which is
    a bug that caused BZERO to be incorrectly set to a value in a compressed
    FITS HDU if a header was passed in with a value even though the data was
    specified as a floating-point value.
    """

    data = np.arange(0, 100, dtype=np.uint16)
    hdu = fits.ImageHDU(data=data)
    data = hdu.data.astype(np.float64)
    header = hdu.header

    hdulist = fits.HDUList(
        [
            fits.PrimaryHDU(),
            fits.ImageHDU(data=data, header=header),
            fits.CompImageHDU(data=data, header=header),
        ]
    )

    hdulist.writeto(tmp_path / "test_bzero.fits")

    for hdu in hdulist:
        assert hdu.header.get("BZERO") is None
        assert hdu.header.get("BSCALE") is None

    with fits.open(tmp_path / "test_bzero.fits") as hdulist_read:
        for hdu in hdulist_read:
            assert hdu.header.get("BZERO") is None
            assert hdu.header.get("BSCALE") is None
        np.testing.assert_array_equal(hdulist_read[1].data, hdulist_read[2].data)


def test_custom_extname():
    """
    Regression test for a bug that caused specifying a custom name for a
    CompImageHDU to not work if an existing image header was used.
    """
    data = np.arange(0, 100, dtype=np.uint16)
    hdu = fits.ImageHDU(data=data)
    header = hdu.header
    fits.CompImageHDU(header=header, name="compressed")


def test_pickle():
    """
    Regression test for https://github.com/astropy/astropy/issues/10512 which
    was a bug that caused CompImageHDU to not be picklable.
    """
    data = np.array([1, 2, 3])
    hdu = fits.CompImageHDU(data)
    a = pickle.loads(pickle.dumps(hdu))
    assert_equal(hdu.data, data)


def test_compression_settings_delayed_data(tmp_path):
    """
    Regression test for https://github.com/astropy/astropy/issues/12216 which
    was a bug that caused compression settings to be ignored if the data was
    only set later.
    """

    hdu1 = fits.CompImageHDU(quantize_level=-32)
    hdu2 = fits.CompImageHDU(quantize_level=-32, data=np.array([0.0]))

    with NumpyRNGContext(42):
        data = np.random.random((16, 16))

    hdu1.data = data
    hdu2.data = data

    hdu1.writeto(tmp_path / "data1.fits")
    hdu2.writeto(tmp_path / "data2.fits")

    with fits.open(tmp_path / "data1.fits") as hdulist1_read:
        with fits.open(tmp_path / "data2.fits") as hdulist2_read:
            assert_equal(hdulist1_read[1].data, hdulist2_read[1].data)


def test_header_assignment_issue(tmp_path):
    """
    Regression test for https://github.com/astropy/astropy/issues/14081 which
    was a bug that caused entries in header to not be preserved under certain
    conditions if copied from another header.
    """

    ih = fits.ImageHDU()
    ih.header["test"] = "right"

    ch = fits.CompImageHDU()
    ch.header["test"] = "wrong"
    ch.header = ih.header
    assert ch.header["test"] == "right"

    ch.writeto(tmp_path / "test_header.fits")
    with fits.open(tmp_path / "test_header.fits") as chdl:
        assert chdl[1].header["test"] == "right"


def test_section_unwritten():
    """
    Regression test for https://github.com/astropy/astropy/issues/14611 which
    was a bug that caused CompImageHDU.section to not work correctly if the
    file was not written out to disk first.
    """

    data = np.arange(21 * 33).reshape((21, 33)).astype(np.int32)
    header = fits.Header()
    hdu = fits.CompImageHDU(data, header, compression_type="RICE_1", tile_shape=(5, 6))
    assert_equal(hdu.section[...], data)
    assert hdu.section[3, 4] == data[3, 4]


def test_section_scaling(tmp_path):
    data = (np.arange(200 * 200).reshape(200, 200) - 20_000).astype(np.int16)
    chdu = fits.CompImageHDU(data=data)
    chdu._scale_internal(bzero=1.234, bscale=0.0002, blank=32767)
    chdu.writeto(tmp_path / "compressed.fits")

    with fits.open(tmp_path / "compressed.fits") as hdul:
        section = hdul[1].section
        # underlying data is int16 but accessing it triggers scaling
        assert section.dtype == np.int16
        assert section[:10].dtype == np.float32
        # behavior should remain the same after data has been accessed
        assert section.dtype == np.int16
        assert section[:10].dtype == np.float32

    with fits.open(tmp_path / "compressed.fits") as hdul:
        section = hdul[1].section
        # underlying data is int16 but accessing it triggers scaling
        assert section.dtype == np.int16
        assert section[...].dtype == np.float32
        # behavior should remain the same after data has been accessed
        assert section.dtype == np.int16
        assert section[...].dtype == np.float32


EXPECTED_HEADER = """
XTENSION= 'IMAGE   '           / Image extension
BITPIX  =                   16 / data type of original image
NAXIS   =                    2 / dimension of original image
NAXIS1  =                   10 / length of original image axis
NAXIS2  =                   10 / length of original image axis
PCOUNT  =                    0 / number of parameters
GCOUNT  =                    1 / number of groups
END
""".lstrip()


def test_header():
    """
    Check that the header is correct when reading in compressed images and
    correctly shows the image dimensions.
    """

    filename = os.path.join(
        os.path.dirname(__file__),
        "..",
        "..",
        "..",
        "tests",
        "data",
        "compressed_image.fits",
    )

    with fits.open(filename) as hdulist:
        assert hdulist[1].header == fits.Header.fromstring(EXPECTED_HEADER, sep="\n")


def test_rice_one_alias():
    # Regression test for a bug that caused RICE_ONE (which we document as an
    # acceptable alias) to no longer be recognized.
    chdu = fits.CompImageHDU(np.zeros((3, 4, 5)))
    chdu.compression_type = "RICE_ONE"


def test_header_order(tmp_path):
    # Make sure that comments and history entries appear in the same order and
    # same location after compression and decompression.

    data = np.random.random((128, 128))
    header = fits.Header()

    header["a"] = 1
    header["b"] = 2
    header["c"] = 3
    header["d"] = 4

    header.add_history("history one", before="a")
    header.add_comment("comment one", before="a")
    header.add_comment("comment two", after="a")
    header.add_comment("comment three", before="b")
    header.add_history("history two", before="b")
    header.add_comment("comment four", after="b")
    header.add_comment("comment five", after="d")
    header.add_history("history three")
    header.add_blank()
    header.add_blank()
    header.add_blank()

    hdulist = fits.HDUList([fits.PrimaryHDU(), fits.CompImageHDU(data, header)])

    hdulist.writeto(tmp_path / "test.fits", overwrite=True)

    with fits.open(tmp_path / "test.fits") as hdulist2:
        assert hdulist[1].header.tostring("\n") == hdulist2[1].header.tostring("\n")


def test_hdu_lazy_loading(tmp_path):
    # Lazy loading of HDUs relies on parameters such as _data_offset and so on,
    # so we need to make sure these correctly point to the internal BinTableHDU.

    hdulist = fits.HDUList([fits.PrimaryHDU()])

    for idx in range(6):
        data = np.ones((2**idx, 2**idx)) * idx
        hdulist.append(fits.CompImageHDU(data))

    hdulist.writeto(tmp_path / "multi_hdu.fits")

    with open(tmp_path / "multi_hdu.fits", "rb") as fileobj:
        with fits.open(fileobj) as hdulist:
            for idx in range(6):
                assert_equal(hdulist[idx + 1].data, idx)
                assert hdulist[idx + 1].data.shape == (2**idx, 2**idx)
                fileobj.seek(0)


@pytest.mark.parametrize(
    "init_kwargs, expected_zflags",
    [
        pytest.param(
            dict(quantize_level=-32, data=np.array([0.0])),
            {
                "ZNAME1": "BLOCKSIZE",
                "ZVAL1": 32,
                "ZNAME2": "BYTEPIX",
                "ZVAL2": 4,
                "ZNAME3": "NOISEBIT",
                "ZVAL3": -32,
                "ZCMPTYPE": "RICE_1",
                "ZQUANTIZ": "NO_DITHER",
            },
            id="quantize_level_w_data",
        ),
        pytest.param(
            dict(quantize_level=-32),
            {
                "ZNAME1": "BLOCKSIZE",
                "ZVAL1": 32,
                "ZNAME2": "BYTEPIX",
                "ZVAL2": 4,
                "ZNAME3": "NOISEBIT",
                "ZVAL3": -32,
                "ZCMPTYPE": "RICE_1",
                "ZQUANTIZ": "NO_DITHER",
            },
            id="quantize_level_wo_data",
        ),
        pytest.param(
            dict(quantize_method=2, dither_seed=1, data=np.array([0.0])),
            {
                "ZNAME1": "BLOCKSIZE",
                "ZVAL1": 32,
                "ZNAME2": "BYTEPIX",
                "ZVAL2": 4,
                "ZNAME3": "NOISEBIT",
                "ZVAL3": 16,
                "ZCMPTYPE": "RICE_1",
                "ZQUANTIZ": "SUBTRACTIVE_DITHER_2",
            },
            id="quantize_level_dither_seed_w_data",
        ),
        pytest.param(
            dict(quantize_method=2, dither_seed=1),
            {
                "ZNAME1": "BLOCKSIZE",
                "ZVAL1": 32,
                "ZNAME2": "BYTEPIX",
                "ZVAL2": 4,
                "ZNAME3": "NOISEBIT",
                "ZVAL3": 16,
                "ZCMPTYPE": "RICE_1",
                "ZQUANTIZ": "SUBTRACTIVE_DITHER_2",
            },
            id="quantize_level_dither_seed_wo_data",
        ),
    ],
)
def test_compression_options_with_mutated_data(
    init_kwargs,
    expected_zflags,
    tmp_path,
):
    # see https://github.com/astropy/astropy/issues/12216
    prng = np.random.default_rng(seed=0)
    data = prng.random((50, 30)).astype("float32")

    hdu = fits.CompImageHDU(**init_kwargs)
    hdu.data = data

    hdu.writeto(tmp_path / "t.fits")

    hdr = fits.getheader(tmp_path / "t.fits", ext=1, disable_image_compression=True)

    zflags = {k: hdr[k] for k in expected_zflags}
    assert zflags == expected_zflags


def test_reserved_keywords_stripped(tmp_path):
    # Regression test for a bug that caused THEAP, ZBLANK, ZSCALE and ZZERO to
    # not be correctly stripped from the compressed header when decompressing
    #
    # See also https://github.com/astropy/astropy/issues/18067

    data = np.arange(6).reshape((2, 3))

    hdu = fits.CompImageHDU(data)
    hdu.writeto(tmp_path / "compressed.fits")

    with fits.open(
        tmp_path / "compressed.fits", disable_image_compression=True
    ) as hduc:
        hduc[1].header["THEAP"] = hduc[1].header["NAXIS1"] * hduc[1].header["NAXIS2"]
        hduc[1].header["ZBLANK"] = 1231212
        hduc[1].header["ZSCALE"] = 2
        hduc[1].header["ZZERO"] = 10
        hduc[1].writeto(tmp_path / "compressed_with_extra.fits")

    with fits.open(tmp_path / "compressed_with_extra.fits") as hdud:
        assert "THEAP" not in hdud[1].header
        assert "ZBLANK" not in hdud[1].header
        assert "ZSCALE" not in hdud[1].header
        assert "ZZERO" not in hdud[1].header



from copy import copy

import pytest
from numpy.testing import assert_equal

from astropy.table import Column, Table
from astropy.table.mixins.registry import (
    MixinRegistryError,
    _handlers,
    get_mixin_handler,
    register_mixin_handler,
)
from astropy.table.table_helpers import ArrayWrapper

ORIGINAL = {}


def setup_function(function):
    ORIGINAL["handlers"] = copy(_handlers)
    _handlers.clear()


def teardown_function(function):
    _handlers.clear()
    _handlers.update(ORIGINAL["handlers"])


class SpamData:
    pass


class SpamWrapper(ArrayWrapper):
    def __init__(self):
        super().__init__([0, 1, 3, 4, 5])


FULL_QUALNAME = "astropy.table.mixins.tests.test_registry.SpamData"


def handle_spam(obj):
    return SpamWrapper()


def handle_spam_alt(obj):
    return SpamWrapper()


def test_no_handler():
    data = SpamData()
    assert get_mixin_handler(data) is None


def test_register_handler():
    register_mixin_handler(FULL_QUALNAME, handle_spam)
    assert get_mixin_handler(SpamData()) is handle_spam


def test_register_handler_override():
    register_mixin_handler(FULL_QUALNAME, handle_spam)
    with pytest.raises(MixinRegistryError) as exc:
        register_mixin_handler(FULL_QUALNAME, handle_spam_alt)
    assert (
        exc.value.args[0]
        == "Handler for class astropy.table.mixins.tests.test_registry.SpamData is"
        " already defined"
    )
    register_mixin_handler(FULL_QUALNAME, handle_spam_alt, force=True)
    assert get_mixin_handler(SpamData()) is handle_spam_alt


def test_get_mixin_handler_str():
    # Check that we can also pass a fully qualified name to get_mixin_handler
    register_mixin_handler(FULL_QUALNAME, handle_spam)
    assert get_mixin_handler(FULL_QUALNAME) is handle_spam


def test_add_column_to_empty_table():
    t = Table()
    t["a"] = SpamData()
    # By default, we get an object column.
    assert isinstance(t["a"], Column)
    assert t["a"].dtype == object
    # But after registration, we can get the intended mixin.
    register_mixin_handler(FULL_QUALNAME, handle_spam)
    t = Table()
    t["a"] = SpamData()

    assert len(t) == 5
    assert isinstance(t["a"], SpamWrapper)
    assert_equal(t["a"].data, [0, 1, 3, 4, 5])


def test_add_column_to_existing_table():
    # As above, but for a table that already has a column
    # (addition used to depend on whether or a table was empty; gh-17102).
    t = Table([[5, 6, 7, 8, 9]], names=["x"])
    t["a"] = SpamData()
    assert isinstance(t["a"], Column)
    assert t["a"].dtype == object

    register_mixin_handler(FULL_QUALNAME, handle_spam)

    t["a"] = SpamData()

    assert len(t) == 5
    assert isinstance(t["a"], SpamWrapper)
    assert_equal(t["a"].data, [0, 1, 3, 4, 5])


def invalid_handler(obj):
    return "invalid"


def test_invalid_handler():
    t = Table()

    register_mixin_handler(FULL_QUALNAME, invalid_handler)

    with pytest.raises(TypeError) as exc:
        t["a"] = SpamData()
    assert (
        exc.value.args[0] == f"Mixin handler for object of type {FULL_QUALNAME} "
        "did not return a valid mixin column"
    )
